@@ -49,7 +49,6 @@ module Hledger.Data.Amount (
   isNonsimpleCommodityChar,
   quoteCommoditySymbolIfNeeded,
   -- * Amount
-  amount,
   nullamt,
   missingamt,
   num,
@@ -67,7 +66,6 @@ module Hledger.Data.Amount (
   amountLooksZero,
   divideAmount,
   multiplyAmount,
-  amountTotalPriceToUnitPrice,
   -- ** rendering
   AmountDisplayOpts(..),
   noColour,
@@ -125,7 +123,6 @@ module Hledger.Data.Amount (
   maIsZero,
   maIsNonZero,
   mixedAmountLooksZero,
-  mixedAmountTotalPriceToUnitPrice,
   -- ** rendering
   styleMixedAmount,
   mixedAmountUnstyled,
@@ -144,11 +141,11 @@ module Hledger.Data.Amount (
   mixedAmountSetFullPrecision,
   canonicaliseMixedAmount,
   -- * misc.
-  ltraceamount,
   tests_Amount
 ) where
 
-import Control.Applicative (liftA2)
+import Prelude hiding (Applicative(..))
+import Control.Applicative (Applicative(..))
 import Control.Monad (foldM)
 import Data.Char (isDigit)
 import Data.Decimal (DecimalRaw(..), decimalPlaces, normalizeDecimal, roundTo)
@@ -166,12 +163,11 @@ import Data.Word (Word8)
 import Safe (headDef, lastDef, lastMay)
 import System.Console.ANSI (Color(..),ColorIntensity(..))
 
-import Debug.Trace (trace)
 import Test.Tasty (testGroup)
 import Test.Tasty.HUnit ((@?=), assertBool, testCase)
 
 import Hledger.Data.Types
-import Hledger.Utils (colorB)
+import Hledger.Utils (colorB, numDigitsInt)
 import Hledger.Utils.Text (textQuoteIfNeeded)
 import Text.WideString (WideBuilder(..), wbFromText, wbToText, wbUnpack)
 
@@ -258,22 +254,22 @@ instance Num Amount where
     (*)                          = similarAmountsOp (*)
 
 -- | The empty simple amount.
-amount, nullamt :: Amount
-amount = Amount{acommodity="", aquantity=0, aprice=Nothing, astyle=amountstyle}
-nullamt = amount
+nullamt :: Amount
+nullamt = Amount{acommodity="", aquantity=0, aprice=Nothing, astyle=amountstyle}
 
 -- | A temporary value for parsed transactions which had no amount specified.
 missingamt :: Amount
-missingamt = amount{acommodity="AUTO"}
+missingamt = nullamt{acommodity="AUTO"}
 
 -- Handy amount constructors for tests.
 -- usd/eur/gbp round their argument to a whole number of pennies/cents.
-num n = amount{acommodity="",  aquantity=n}
-hrs n = amount{acommodity="h", aquantity=n,           astyle=amountstyle{asprecision=Precision 2, ascommodityside=R}}
-usd n = amount{acommodity="$", aquantity=roundTo 2 n, astyle=amountstyle{asprecision=Precision 2}}
-eur n = amount{acommodity="€", aquantity=roundTo 2 n, astyle=amountstyle{asprecision=Precision 2}}
-gbp n = amount{acommodity="£", aquantity=roundTo 2 n, astyle=amountstyle{asprecision=Precision 2}}
-per n = amount{acommodity="%", aquantity=n,           astyle=amountstyle{asprecision=Precision 1, ascommodityside=R, ascommodityspaced=True}}
+-- XXX these are a bit clashy
+num n = nullamt{acommodity="",  aquantity=n}
+hrs n = nullamt{acommodity="h", aquantity=n,           astyle=amountstyle{asprecision=Precision 2, ascommodityside=R}}
+usd n = nullamt{acommodity="$", aquantity=roundTo 2 n, astyle=amountstyle{asprecision=Precision 2}}
+eur n = nullamt{acommodity="€", aquantity=roundTo 2 n, astyle=amountstyle{asprecision=Precision 2}}
+gbp n = nullamt{acommodity="£", aquantity=roundTo 2 n, astyle=amountstyle{asprecision=Precision 2}}
+per n = nullamt{acommodity="%", aquantity=n,           astyle=amountstyle{asprecision=Precision 1, ascommodityside=R, ascommodityspaced=True}}
 amt `at` priceamt = amt{aprice=Just $ UnitPrice priceamt}
 amt @@ priceamt = amt{aprice=Just $ TotalPrice priceamt}
 
@@ -288,7 +284,7 @@ similarAmountsOp :: (Quantity -> Quantity -> Quantity) -> Amount -> Amount -> Am
 similarAmountsOp op Amount{acommodity=_,  aquantity=q1, astyle=AmountStyle{asprecision=p1}}
                     Amount{acommodity=c2, aquantity=q2, astyle=s2@AmountStyle{asprecision=p2}} =
    -- trace ("a1:"++showAmountDebug a1) $ trace ("a2:"++showAmountDebug a2) $ traceWith (("= :"++).showAmountDebug)
-   amount{acommodity=c2, aquantity=q1 `op` q2, astyle=s2{asprecision=max p1 p2}}
+   nullamt{acommodity=c2, aquantity=q1 `op` q2, astyle=s2{asprecision=max p1 p2}}
   --  c1==c2 || q1==0 || q2==0 =
   --  otherwise = error "tried to do simple arithmetic with amounts in different commodities"
 
@@ -312,28 +308,12 @@ amountCost a@Amount{aquantity=q, aprice=mp} =
       Just (UnitPrice  p@Amount{aquantity=pq}) -> p{aquantity=pq * q}
       Just (TotalPrice p@Amount{aquantity=pq}) -> p{aquantity=pq}
 
--- | Replace an amount's TotalPrice, if it has one, with an equivalent UnitPrice.
--- Has no effect on amounts without one.
--- Also increases the unit price's display precision to show one extra decimal place,
--- to help keep transaction amounts balancing.
--- Does Decimal division, might be some rounding/irrational number issues.
-amountTotalPriceToUnitPrice :: Amount -> Amount
-amountTotalPriceToUnitPrice
-    a@Amount{aquantity=q, aprice=Just (TotalPrice pa@Amount{aquantity=pq, astyle=ps})}
-    = a{aprice = Just $ UnitPrice pa{aquantity=abs (pq/q), astyle=ps{asprecision=pp}}}
-  where
-    -- Increase the precision by 1, capping at the max bound.
-    pp = case asprecision ps of
-                NaturalPrecision -> NaturalPrecision
-                Precision p      -> Precision $ if p == maxBound then maxBound else p + 1
-amountTotalPriceToUnitPrice a = a
-
 -- | Apply a function to an amount's quantity (and its total price, if it has one).
 transformAmount :: (Quantity -> Quantity) -> Amount -> Amount
 transformAmount f a@Amount{aquantity=q,aprice=p} = a{aquantity=f q, aprice=f' <$> p}
   where
-    f' (TotalPrice a@Amount{aquantity=pq}) = TotalPrice a{aquantity = f pq}
-    f' p = p
+    f' (TotalPrice a1@Amount{aquantity=pq}) = TotalPrice a1{aquantity = f pq}
+    f' p' = p'
 
 -- | Divide an amount's quantity (and its total price, if it has one) by a constant.
 divideAmount :: Quantity -> Amount -> Amount
@@ -540,15 +520,15 @@ showamountquantity amt@Amount{astyle=AmountStyle{asdecimalpoint=mdec, asdigitgro
 applyDigitGroupStyle :: Maybe DigitGroupStyle -> Int -> T.Text -> WideBuilder
 applyDigitGroupStyle Nothing                       l s = WideBuilder (TB.fromText s) l
 applyDigitGroupStyle (Just (DigitGroups _ []))     l s = WideBuilder (TB.fromText s) l
-applyDigitGroupStyle (Just (DigitGroups c (g:gs))) l s = addseps (g:|gs) (toInteger l) s
+applyDigitGroupStyle (Just (DigitGroups c (g0:gs0))) l0 s0 = addseps (g0:|gs0) (toInteger l0) s0
   where
-    addseps (g:|gs) l s
-        | l' > 0    = addseps gs' l' rest <> WideBuilder (TB.singleton c <> TB.fromText part) (fromIntegral g + 1)
-        | otherwise = WideBuilder (TB.fromText s) (fromInteger l)
+    addseps (g1:|gs1) l1 s1
+        | l2 > 0    = addseps gs2 l2 rest <> WideBuilder (TB.singleton c <> TB.fromText part) (fromIntegral g1 + 1)
+        | otherwise = WideBuilder (TB.fromText s1) (fromInteger l1)
       where
-        (rest, part) = T.splitAt (fromInteger l') s
-        gs' = fromMaybe (g:|[]) $ nonEmpty gs
-        l' = l - toInteger g
+        (rest, part) = T.splitAt (fromInteger l2) s1
+        gs2 = fromMaybe (g1:|[]) $ nonEmpty gs1
+        l2 = l1 - toInteger g1
 
 -- like journalCanonicaliseAmounts
 -- | Canonicalise an amount's display style using the provided commodity style map.
@@ -604,10 +584,12 @@ mixedAmount :: Amount -> MixedAmount
 mixedAmount a = Mixed $ M.singleton (amountKey a) a
 
 -- | Add an Amount to a MixedAmount, normalising the result.
+-- Amounts with different costs are kept separate.
 maAddAmount :: MixedAmount -> Amount -> MixedAmount
 maAddAmount (Mixed ma) a = Mixed $ M.insertWith sumSimilarAmountsUsingFirstPrice (amountKey a) a ma
 
 -- | Add a collection of Amounts to a MixedAmount, normalising the result.
+-- Amounts with different costs are kept separate.
 maAddAmounts :: Foldable t => MixedAmount -> t Amount -> MixedAmount
 maAddAmounts = foldl' maAddAmount
 
@@ -615,15 +597,18 @@ maAddAmounts = foldl' maAddAmount
 maNegate :: MixedAmount -> MixedAmount
 maNegate = transformMixedAmount negate
 
--- | Sum two MixedAmount.
+-- | Sum two MixedAmount, keeping the cost of the first if any.
+-- Amounts with different costs are kept separate (since 2021).
 maPlus :: MixedAmount -> MixedAmount -> MixedAmount
 maPlus (Mixed as) (Mixed bs) = Mixed $ M.unionWith sumSimilarAmountsUsingFirstPrice as bs
 
 -- | Subtract a MixedAmount from another.
+-- Amounts with different costs are kept separate.
 maMinus :: MixedAmount -> MixedAmount -> MixedAmount
 maMinus a = maPlus a . maNegate
 
 -- | Sum a collection of MixedAmounts.
+-- Amounts with different costs are kept separate.
 maSum :: Foldable t => t MixedAmount -> MixedAmount
 maSum = foldl' maPlus nullmixedamt
 
@@ -720,11 +705,11 @@ maCommodities = S.fromList . fmap acommodity . amounts'
 unifyMixedAmount :: MixedAmount -> Maybe Amount
 unifyMixedAmount = foldM combine 0 . amounts
   where
-    combine amount result
-      | amountIsZero amount                    = Just result
-      | amountIsZero result                    = Just amount
-      | acommodity amount == acommodity result = Just $ amount + result
-      | otherwise                              = Nothing
+    combine amt result
+      | amountIsZero amt                    = Just result
+      | amountIsZero result                 = Just amt
+      | acommodity amt == acommodity result = Just $ amt + result
+      | otherwise                           = Nothing
 
 -- | Sum same-commodity amounts in a lossy way, applying the first
 -- price to the result and discarding any other prices. Only used as a
@@ -800,11 +785,11 @@ mixedAmountUnstyled = mapMixedAmountUnsafe amountUnstyled
 showMixedAmount :: MixedAmount -> String
 showMixedAmount = wbUnpack . showMixedAmountB noColour
 
--- | Get the one-line string representation of a mixed amount.
+-- | Get the one-line string representation of a mixed amount (also showing any costs).
 --
 -- > showMixedAmountOneLine = wbUnpack . showMixedAmountB oneLine
 showMixedAmountOneLine :: MixedAmount -> String
-showMixedAmountOneLine = wbUnpack . showMixedAmountB oneLine
+showMixedAmountOneLine = wbUnpack . showMixedAmountB oneLine{displayPrice=True}
 
 -- | Like showMixedAmount, but zero amounts are shown with their
 -- commodity if they have one.
@@ -857,10 +842,10 @@ showMixedAmountDebug m | m == missingmixedamt = "(missing)"
 showMixedAmountB :: AmountDisplayOpts -> MixedAmount -> WideBuilder
 showMixedAmountB opts ma
     | displayOneLine opts = showMixedAmountOneLineB opts ma
-    | otherwise           = WideBuilder (wbBuilder . mconcat $ intersperse sep lines) width
+    | otherwise           = WideBuilder (wbBuilder . mconcat $ intersperse sep ls) width
   where
-    lines = showMixedAmountLinesB opts ma
-    width = headDef 0 $ map wbWidth lines
+    ls = showMixedAmountLinesB opts ma
+    width = headDef 0 $ map wbWidth ls
     sep = WideBuilder (TB.singleton '\n') 0
 
 -- | Helper for showMixedAmountB to show a list of Amounts on multiple lines. This returns
@@ -918,7 +903,7 @@ showMixedAmountOneLineB opts@AmountDisplayOpts{displayMaxWidth=mmax,displayMinWi
     dropWhileRev p = foldr (\x xs -> if null xs && p x then [] else x:xs) []
 
     -- Add the elision strings (if any) to each amount
-    withElided = zipWith (\num amt -> (amt, elisionDisplay Nothing (wbWidth sep) num amt)) [n-1,n-2..0]
+    withElided = zipWith (\n2 amt -> (amt, elisionDisplay Nothing (wbWidth sep) n2 amt)) [n-1,n-2..0]
 
 orderedAmounts :: AmountDisplayOpts -> MixedAmount -> [Amount]
 orderedAmounts dopts = maybe id (mapM pad) (displayOrder dopts) . amounts
@@ -950,8 +935,8 @@ elisionDisplay mmax sep n lastAmt
   | otherwise = Nothing
   where
     fullString = T.pack $ show n ++ " more.."
-    -- sep from the separator, 7 from " more..", 1 + floor (logBase 10 n) from number
-    fullLength = sep + 8 + floor (logBase 10 $ fromIntegral n)
+    -- sep from the separator, 7 from " more..", numDigits n from number
+    fullLength = sep + 7 + numDigitsInt n
 
     str | Just m <- mmax, fullLength > m = T.take (m - 2) fullString <> ".."
         | otherwise                      = fullString
@@ -962,10 +947,6 @@ maybeAppend :: Maybe a -> [a] -> [a]
 maybeAppend Nothing  = id
 maybeAppend (Just a) = (++[a])
 
--- | Compact labelled trace of a mixed amount, for debugging.
-ltraceamount :: String -> MixedAmount -> MixedAmount
-ltraceamount s a = trace (s ++ ": " ++ showMixedAmount a) a
-
 -- | Set the display precision in the amount's commodities.
 mixedAmountSetPrecision :: AmountPrecision -> MixedAmount -> MixedAmount
 mixedAmountSetPrecision p = mapMixedAmountUnsafe (amountSetPrecision p)
@@ -985,12 +966,6 @@ mixedAmountStripPrices (Mixed ma) =
 canonicaliseMixedAmount :: M.Map CommoditySymbol AmountStyle -> MixedAmount -> MixedAmount
 canonicaliseMixedAmount styles = mapMixedAmountUnsafe (canonicaliseAmount styles)
 
--- | Replace each component amount's TotalPrice, if it has one, with an equivalent UnitPrice.
--- Has no effect on amounts without one.
--- Does Decimal division, might be some rounding/irrational number issues.
-mixedAmountTotalPriceToUnitPrice :: MixedAmount -> MixedAmount
-mixedAmountTotalPriceToUnitPrice = mapMixedAmount amountTotalPriceToUnitPrice
-
 
 -------------------------------------------------------------------------------
 -- tests
@@ -1005,7 +980,7 @@ tests_Amount = testGroup "Amount" [
        amountCost (eur (-1)){aprice=Just $ TotalPrice $ usd (-2)} @?= usd (-2)
 
     ,testCase "amountLooksZero" $ do
-       assertBool "" $ amountLooksZero amount
+       assertBool "" $ amountLooksZero nullamt
        assertBool "" $ amountLooksZero $ usd 0
 
     ,testCase "negating amounts" $ do
