@@ -18,7 +18,6 @@ module Hledger.Data.Balancing
 , isTransactionBalanced
 , balanceTransaction
 , balanceTransactionHelper
-, annotateErrorWithTransaction
   -- * journal balancing
 , journalBalanceTransactions
 , journalCheckBalanceAssertions
@@ -27,16 +26,20 @@ module Hledger.Data.Balancing
 )
 where
 
+import Control.Monad (forM, forM_, when, unless)
 import Control.Monad.Except (ExceptT(..), runExceptT, throwError)
 import "extra" Control.Monad.Extra (whenM)
-import Control.Monad.Reader as R
+import Control.Monad.Reader as R (ReaderT, reader, runReaderT, ask, asks)
 import Control.Monad.ST (ST, runST)
+import Control.Monad.Trans.Class (lift)
 import Data.Array.ST (STArray, getElems, newListArray, writeArray)
 import Data.Foldable (asum)
 import Data.Function ((&))
+import Data.Functor ((<&>))
+import "base-compat" Data.Functor.Compat (void)
 import qualified Data.HashTable.Class as H (toList)
 import qualified Data.HashTable.ST.Cuckoo as H
-import Data.List (intercalate, partition, sortOn)
+import Data.List (partition, sortOn)
 import Data.List.Extra (nubSort)
 import Data.Maybe (fromJust, fromMaybe, isJust, isNothing, mapMaybe)
 import qualified Data.Set as S
@@ -50,23 +53,24 @@ import Hledger.Utils
 import Hledger.Data.Types
 import Hledger.Data.AccountName (isAccountNamePrefixOf)
 import Hledger.Data.Amount
-import Hledger.Data.Dates (showDate)
 import Hledger.Data.Journal
 import Hledger.Data.Posting
 import Hledger.Data.Transaction
+import Hledger.Data.Errors
 
 
 data BalancingOpts = BalancingOpts
-  { ignore_assertions_        :: Bool  -- ^ Ignore balance assertions
-  , infer_transaction_prices_ :: Bool  -- ^ Infer prices in unbalanced multicommodity amounts
-  , commodity_styles_         :: Maybe (M.Map CommoditySymbol AmountStyle)  -- ^ commodity display styles
+  { ignore_assertions_     :: Bool  -- ^ should failing balance assertions be ignored ?
+  , infer_balancing_costs_ :: Bool  -- ^ Are we permitted to infer missing costs to balance transactions ?
+                                    --   Distinct from InputOpts{infer_costs_}.
+  , commodity_styles_      :: Maybe (M.Map CommoditySymbol AmountStyle)  -- ^ commodity display styles
   } deriving (Show)
 
 defbalancingopts :: BalancingOpts
 defbalancingopts = BalancingOpts
-  { ignore_assertions_        = False
-  , infer_transaction_prices_ = True
-  , commodity_styles_         = Nothing
+  { ignore_assertions_     = False
+  , infer_balancing_costs_ = True
+  , commodity_styles_      = Nothing
   }
 
 -- | Check that this transaction would appear balanced to a human when displayed.
@@ -98,16 +102,18 @@ transactionCheckBalanced BalancingOpts{commodity_styles_} t = errs
 
     -- check for mixed signs, detecting nonzeros at display precision
     canonicalise = maybe id canonicaliseMixedAmount commodity_styles_
+    postingBalancingAmount p
+      | "_price-matched" `elem` map fst (ptags p) = mixedAmountStripPrices $ pamount p
+      | otherwise                                 = mixedAmountCost $ pamount p
     signsOk ps =
-      case filter (not.mixedAmountLooksZero) $ map (canonicalise.mixedAmountCost.pamount) ps of
+      case filter (not.mixedAmountLooksZero) $ map (canonicalise.postingBalancingAmount) ps of
         nonzeros | length nonzeros >= 2
                    -> length (nubSort $ mapMaybe isNegativeMixedAmount nonzeros) > 1
         _          -> True
     (rsignsok, bvsignsok)       = (signsOk rps, signsOk bvps)
 
     -- check for zero sum, at display precision
-    (rsum, bvsum)               = (sumPostings rps, sumPostings bvps)
-    (rsumcost, bvsumcost)       = (mixedAmountCost rsum, mixedAmountCost bvsum)
+    (rsumcost, bvsumcost)       = (foldMap postingBalancingAmount rps, foldMap postingBalancingAmount bvps)
     (rsumdisplay, bvsumdisplay) = (canonicalise rsumcost, canonicalise bvsumcost)
     (rsumok, bvsumok)           = (mixedAmountLooksZero rsumdisplay, mixedAmountLooksZero bvsumdisplay)
 
@@ -116,12 +122,12 @@ transactionCheckBalanced BalancingOpts{commodity_styles_} t = errs
       where
         rmsg
           | rsumok        = ""
-          | not rsignsok  = "real postings all have the same sign"
-          | otherwise     = "real postings' sum should be 0 but is: " ++ showMixedAmount rsumcost
+          | not rsignsok  = "The real postings all have the same sign. Consider negating some of them."
+          | otherwise     = "The real postings' sum should be 0 but is: " ++ showMixedAmountOneLineWithoutPrice False rsumcost
         bvmsg
           | bvsumok       = ""
-          | not bvsignsok = "balanced virtual postings all have the same sign"
-          | otherwise     = "balanced virtual postings' sum should be 0 but is: " ++ showMixedAmount bvsumcost
+          | not bvsignsok = "The balanced virtual postings all have the same sign. Consider negating some of them."
+          | otherwise     = "The balanced virtual postings' sum should be 0 but is: " ++ showMixedAmountOneLineWithoutPrice False bvsumcost
 
 -- | Legacy form of transactionCheckBalanced.
 isTransactionBalanced :: BalancingOpts -> Transaction -> Bool
@@ -147,31 +153,55 @@ balanceTransaction ::
 balanceTransaction bopts = fmap fst . balanceTransactionHelper bopts
 
 -- | Helper used by balanceTransaction and balanceTransactionWithBalanceAssignmentAndCheckAssertionsB;
--- use one of those instead. It also returns a list of accounts
--- and amounts that were inferred.
+-- use one of those instead.
+-- It also returns a list of accounts and amounts that were inferred.
 balanceTransactionHelper ::
      BalancingOpts
   -> Transaction
   -> Either String (Transaction, [(AccountName, MixedAmount)])
 balanceTransactionHelper bopts t = do
-  (t', inferredamtsandaccts) <- inferBalancingAmount (fromMaybe M.empty $ commodity_styles_ bopts) $
-    if infer_transaction_prices_ bopts then inferBalancingPrices t else t
+  (t', inferredamtsandaccts) <-
+    transactionInferBalancingAmount (fromMaybe M.empty $ commodity_styles_ bopts) $
+    (if infer_balancing_costs_ bopts then transactionInferBalancingCosts else id)
+    t
   case transactionCheckBalanced bopts t' of
     []   -> Right (txnTieKnot t', inferredamtsandaccts)
-    errs -> Left $ transactionBalanceError t' errs
+    errs -> Left $ transactionBalanceError t' errs'
+      where
+        ismulticommodity = (length $ transactionCommodities t') > 1
+        errs' =
+          [ "Automatic commodity conversion is not enabled."
+          | ismulticommodity && not (infer_balancing_costs_ bopts)
+          ] ++
+          errs ++
+          if ismulticommodity
+          then
+          [ "Consider adjusting this entry's amounts, adding missing postings,"
+          , "or recording conversion price(s) with @, @@ or equity postings." 
+          ]
+          else
+          [ "Consider adjusting this entry's amounts, or adding missing postings."
+          ]
+
+transactionCommodities :: Transaction -> S.Set CommoditySymbol
+transactionCommodities t = mconcat $ map (maCommodities . pamount) $ tpostings t
 
 -- | Generate a transaction balancing error message, given the transaction
 -- and one or more suberror messages.
 transactionBalanceError :: Transaction -> [String] -> String
-transactionBalanceError t errs =
-  annotateErrorWithTransaction t $
-  intercalate "\n" $ "could not balance this transaction:" : errs
-
-annotateErrorWithTransaction :: Transaction -> String -> String
-annotateErrorWithTransaction t s =
-  unlines [ showSourcePosPair $ tsourcepos t, s
-          , T.unpack . T.stripEnd $ showTransaction t
-          ]
+transactionBalanceError t errs = printf "%s:\n%s\n\nThis %stransaction is unbalanced.\n%s"
+  (sourcePosPairPretty $ tsourcepos t)
+  (textChomp ex)
+  (if ismulticommodity then "multi-commodity " else "" :: String)
+  (chomp $ unlines errs)
+  where
+    ismulticommodity = (length $ transactionCommodities t) > 1
+    (_f,_l,_mcols,ex) = makeTransactionErrorExcerpt t finderrcols
+      where
+        finderrcols _ = Nothing
+        -- finderrcols t = Just (1, Just w)
+        --   where
+        --     w = maximumDef 1 $ map T.length $ T.lines $ showTransaction t
 
 -- | Infer up to one missing amount for this transactions's real postings, and
 -- likewise for its balanced virtual postings, if needed; or return an error
@@ -181,19 +211,19 @@ annotateErrorWithTransaction t s =
 -- We can infer a missing amount when there are multiple postings and exactly
 -- one of them is amountless. If the amounts had price(s) the inferred amount
 -- have the same price(s), and will be converted to the price commodity.
-inferBalancingAmount ::
+transactionInferBalancingAmount ::
      M.Map CommoditySymbol AmountStyle -- ^ commodity display styles
   -> Transaction
   -> Either String (Transaction, [(AccountName, MixedAmount)])
-inferBalancingAmount styles t@Transaction{tpostings=ps}
+transactionInferBalancingAmount styles t@Transaction{tpostings=ps}
   | length amountlessrealps > 1
       = Left $ transactionBalanceError t
-        ["can't have more than one real posting with no amount"
-        ,"(remember to put two or more spaces between account and amount)"]
+        ["There can't be more than one real posting with no amount."
+        ,"(Remember to put two or more spaces between account and amount.)"]
   | length amountlessbvps > 1
       = Left $ transactionBalanceError t
-        ["can't have more than one balanced virtual posting with no amount"
-        ,"(remember to put two or more spaces between account and amount)"]
+        ["There can't be more than one balanced virtual posting with no amount."
+        ,"(Remember to put two or more spaces between account and amount.)"]
   | otherwise
       = let psandinferredamts = map inferamount ps
             inferredacctsandamts = [(paccount p, amt) | (p, Just amt) <- psandinferredamts]
@@ -210,6 +240,7 @@ inferBalancingAmount styles t@Transaction{tpostings=ps}
         minferredamt = case ptype p of
           RegularPosting         | not (hasAmount p) -> Just realsum
           BalancedVirtualPosting | not (hasAmount p) -> Just bvsum
+          VirtualPosting         | not (hasAmount p) -> Just 0
           _                                          -> Nothing
       in
         case minferredamt of
@@ -221,55 +252,55 @@ inferBalancingAmount styles t@Transaction{tpostings=ps}
               -- (since the main amount styling pass happened before this balancing pass);
               a' = styleMixedAmount styles . mixedAmountCost $ maNegate a
 
--- | Infer prices for this transaction's posting amounts, if needed to make
--- the postings balance, and if possible. This is done once for the real
+-- | Infer costs for this transaction's posting amounts, if needed to make
+-- the postings balance, and if permitted. This is done once for the real
 -- postings and again (separately) for the balanced virtual postings. When
 -- it's not possible, the transaction is left unchanged.
 --
 -- The simplest example is a transaction with two postings, each in a
--- different commodity, with no prices specified. In this case we'll add a
--- price to the first posting such that it can be converted to the commodity
+-- different commodity, with no costs specified. In this case we'll add a
+-- cost to the first posting such that it can be converted to the commodity
 -- of the second posting (with -B), and such that the postings balance.
 --
--- In general, we can infer a conversion price when the sum of posting amounts
--- contains exactly two different commodities and no explicit prices.  Also
+-- In general, we can infer a cost (conversion rate) when the sum of posting amounts
+-- contains exactly two different commodities and no explicit costs.  Also
 -- all postings are expected to contain an explicit amount (no missing
--- amounts) in a single commodity. Otherwise no price inferring is attempted.
+-- amounts) in a single commodity. Otherwise no cost inferring is attempted.
 --
 -- The transaction itself could contain more than two commodities, and/or
--- prices, if they cancel out; what matters is that the sum of posting amounts
--- contains exactly two commodities and zero prices.
+-- costs, if they cancel out; what matters is that the sum of posting amounts
+-- contains exactly two commodities and zero costs.
 --
 -- There can also be more than two postings in either of the commodities.
 --
--- We want to avoid excessive display of digits when the calculated price is
+-- We want to avoid excessive display of digits when the calculated cost is
 -- an irrational number, while hopefully also ensuring the displayed numbers
 -- make sense if the user does a manual calculation. This is (mostly) achieved
 -- in two ways:
 --
--- - when there is only one posting in the "from" commodity, a total price
+-- - when there is only one posting in the "from" commodity, a total cost
 --   (@@) is used, and all available decimal digits are shown
 --
--- - otherwise, a suitable averaged unit price (@) is applied to the relevant
+-- - otherwise, a suitable averaged unit cost (@) is applied to the relevant
 --   postings, with display precision equal to the summed display precisions
 --   of the two commodities being converted between, or 2, whichever is larger.
 --
--- (We don't always calculate a good-looking display precision for unit prices
+-- (We don't always calculate a good-looking display precision for unit costs
 -- when the commodity display precisions are low, eg when a journal doesn't
--- use any decimal places. The minimum of 2 helps make the prices shown by the
+-- use any decimal places. The minimum of 2 helps make the costs shown by the
 -- print command a bit less surprising in this case. Could do better.)
 --
-inferBalancingPrices :: Transaction -> Transaction
-inferBalancingPrices t@Transaction{tpostings=ps} = t{tpostings=ps'}
+transactionInferBalancingCosts :: Transaction -> Transaction
+transactionInferBalancingCosts t@Transaction{tpostings=ps} = t{tpostings=ps'}
   where
-    ps' = map (priceInferrerFor t BalancedVirtualPosting . priceInferrerFor t RegularPosting) ps
+    ps' = map (costInferrerFor t BalancedVirtualPosting . costInferrerFor t RegularPosting) ps
 
--- | Generate a posting update function which assigns a suitable balancing
--- price to the posting, if and as appropriate for the given transaction and
--- posting type (real or balanced virtual). If we cannot or should not infer
--- prices, just act as the identity on postings.
-priceInferrerFor :: Transaction -> PostingType -> (Posting -> Posting)
-priceInferrerFor t pt = maybe id inferprice inferFromAndTo
+-- | Generate a posting update function which assigns a suitable cost to
+-- balance the posting, if and as appropriate for the given transaction and
+-- posting type (real or balanced virtual) (or if we cannot or should not infer
+-- costs, leaves the posting unchanged).
+costInferrerFor :: Transaction -> PostingType -> (Posting -> Posting)
+costInferrerFor t pt = maybe id infercost inferFromAndTo
   where
     postings     = filter ((==pt).ptype) $ tpostings t
     pcommodities = map acommodity $ concatMap (amounts . pamount) postings
@@ -291,12 +322,12 @@ priceInferrerFor t pt = maybe id inferprice inferFromAndTo
 
     -- For each posting, if the posting type matches, there is only a single amount in the posting,
     -- and the commodity of the amount matches the amount we're converting from,
-    -- then set its price based on the ratio between fromamount and toamount.
-    inferprice (fromamount, toamount) posting
-        | [a] <- amounts (pamount posting), ptype posting == pt, acommodity a == acommodity fromamount
-            = posting{ pamount   = mixedAmount a{aprice=Just conversionprice}
-                     , poriginal = Just $ originalPosting posting }
-        | otherwise = posting
+    -- then set its cost based on the ratio between fromamount and toamount.
+    infercost (fromamount, toamount) p
+        | [a] <- amounts (pamount p), ptype p == pt, acommodity a == acommodity fromamount
+            = p{ pamount   = mixedAmount a{aprice=Just conversionprice}
+                     , poriginal = Just $ originalPosting p }
+        | otherwise = p
       where
         -- If only one Amount in the posting list matches fromamount we can use TotalPrice.
         -- Otherwise divide the conversion equally among the Amounts by using a unit price.
@@ -404,10 +435,11 @@ updateTransactionB :: Transaction -> Balancing s ()
 updateTransactionB t = withRunningBalance $ \BalancingState{bsTransactions}  ->
   void $ writeArray bsTransactions (tindex t) t
 
--- | Infer any missing amounts (to satisfy balance assignments and
--- to balance transactions) and check that all transactions balance
--- and (optional) all balance assertions pass. Or return an error message
--- (just the first error encountered).
+-- | Infer any missing amounts and/or conversion costs
+-- (as needed to balance transactions and satisfy balance assignments);
+-- and check that all transactions are balanced;
+-- and (optional) check that all balance assertions pass.
+-- Or, return an error message (just the first error encountered).
 --
 -- Assumes journalInferCommodityStyles has been called, since those
 -- affect transaction balancing.
@@ -425,18 +457,18 @@ journalBalanceTransactions bopts' j' =
     -- balance assignments are not allowed on accounts affected by auto postings
     autopostingaccts = S.fromList . map (paccount . tmprPosting) . concatMap tmpostingrules $ jtxnmodifiers j
   in
+    -- Store the transactions in a mutable array, which we'll update as we balance them.
+    -- Not strictly necessary but avoids a sort at the end I think.
     runST $ do
-      -- We'll update a mutable array of transactions as we balance them,
-      -- not strictly necessary but avoids a sort at the end I think.
       balancedtxns <- newListArray (1, toInteger $ length ts) ts
 
-      -- Infer missing posting amounts, check transactions are balanced,
-      -- and check balance assertions. This is done in two passes:
+      -- Process all transactions, or short-circuit with an error.
       runExceptT $ do
 
-        -- 1. Step through the transactions, balancing the ones which don't have balance assignments
-        -- and leaving the others for later. The balanced ones are split into their postings.
-        -- The postings and not-yet-balanced transactions remain in the same relative order.
+        -- Two passes are required:
+        -- 1. Step through the transactions, balancing the ones which don't have balance assignments,
+        -- postponing those which do until later. The balanced ones are split into their postings,
+        -- keeping these and the not-yet-balanced transactions in the same relative order.
         psandts :: [Either Posting Transaction] <- fmap concat $ forM ts $ \case
           t | null $ assignmentPostings t -> case balanceTransaction bopts t of
               Left  e  -> throwError e
@@ -445,14 +477,16 @@ journalBalanceTransactions bopts' j' =
                 return $ map Left $ tpostings t'
           t -> return [Right t]
 
-        -- 2. Sort these items by date, preserving the order of same-day items,
-        -- and step through them while keeping running account balances,
+        -- 2. Step through these items in date order (and preserved same-day order),
+        -- keeping running balances for all accounts.
         runningbals <- lift $ H.newSized (length $ journalAccountNamesUsed j)
         flip runReaderT (BalancingState styles autopostingaccts (not $ ignore_assertions_ bopts) runningbals balancedtxns) $ do
-          -- performing balance assignments in, and balancing, the remaining transactions,
-          -- and checking balance assertions as each posting is processed.
+          -- On encountering any not-yet-balanced transaction with a balance assignment,
+          -- enact the balance assignment then finish balancing the transaction.
+          -- And, check any balance assertions encountered along the way.
           void $ mapM' balanceTransactionAndCheckAssertionsB $ sortOn (either postingDate tdate) psandts
 
+        -- Return the now fully-balanced and checked transactions.
         ts' <- lift $ getElems balancedtxns
         return j{jtxns=ts'}
 
@@ -472,9 +506,17 @@ balanceTransactionAndCheckAssertionsB (Left p@Posting{}) =
 balanceTransactionAndCheckAssertionsB (Right t@Transaction{tpostings=ps}) = do
   -- make sure we can handle the balance assignments
   mapM_ checkIllegalBalanceAssignmentB ps
-  -- for each posting, infer its amount from the balance assignment if applicable,
-  -- update the account's running balance and check the balance assertion if any
-  ps' <- mapM (addOrAssignAmountAndCheckAssertionB . postingStripPrices) ps
+  -- for each posting, in date order (though without disturbing their display order),
+  -- 1. infer its amount from the balance assignment if applicable,
+  -- 2. update the account's running balance, and
+  -- 3. check the balance assertion if any.
+  ps' <- ps
+    & zip [1..]                 -- attach original positions
+    & sortOn (postingDate.snd)  -- sort by date
+    & mapM (addOrAssignAmountAndCheckAssertionB)  -- infer amount, check assertion on each one
+    <&> sortOn fst              -- restore original order
+    <&> map snd                 -- discard positions
+
   -- infer any remaining missing amounts, and make sure the transaction is now fully balanced
   styles <- R.reader bsStyles
   case balanceTransactionHelper defbalancingopts{commodity_styles_=styles} t{tpostings=ps'} of
@@ -485,18 +527,20 @@ balanceTransactionAndCheckAssertionsB (Right t@Transaction{tpostings=ps}) = do
       -- and save the balanced transaction.
       updateTransactionB t'
 
+type NumberedPosting = (Integer, Posting)
+
 -- | If this posting has an explicit amount, add it to the account's running balance.
 -- If it has a missing amount and a balance assignment, infer the amount from, and
 -- reset the running balance to, the assigned balance.
 -- If it has a missing amount and no balance assignment, leave it for later.
 -- Then test the balance assertion if any.
-addOrAssignAmountAndCheckAssertionB :: Posting -> Balancing s Posting
-addOrAssignAmountAndCheckAssertionB p@Posting{paccount=acc, pamount=amt, pbalanceassertion=mba}
+addOrAssignAmountAndCheckAssertionB :: NumberedPosting -> Balancing s NumberedPosting
+addOrAssignAmountAndCheckAssertionB (i,p@Posting{paccount=acc, pamount=amt, pbalanceassertion=mba})
   -- an explicit posting amount
   | hasAmount p = do
       newbal <- addToRunningBalanceB acc amt
       whenM (R.reader bsAssrt) $ checkBalanceAssertionB p newbal
-      return p
+      return (i,p)
 
   -- no explicit posting amount, but there is a balance assignment
   | Just BalanceAssertion{baamount,batotal,bainclusive} <- mba = do
@@ -510,10 +554,10 @@ addOrAssignAmountAndCheckAssertionB p@Posting{paccount=acc, pamount=amt, pbalanc
       diff <- (if bainclusive then setInclusiveRunningBalanceB else setRunningBalanceB) acc newbal
       let p' = p{pamount=filterMixedAmount (not . amountIsZero) diff, poriginal=Just $ originalPosting p}
       whenM (R.reader bsAssrt) $ checkBalanceAssertionB p' newbal
-      return p'
+      return (i,p')
 
   -- no explicit posting amount, no balance assignment
-  | otherwise = return p
+  | otherwise = return (i,p)
 
 -- | Add the posting's amount to its account's running balance, and
 -- optionally check the posting's balance assertion if any.
@@ -550,6 +594,7 @@ checkBalanceAssertionB _ _ = return ()
 checkBalanceAssertionOneCommodityB :: Posting -> Amount -> MixedAmount -> Balancing s ()
 checkBalanceAssertionOneCommodityB p@Posting{paccount=assertedacct} assertedamt actualbal = do
   let isinclusive = maybe False bainclusive $ pbalanceassertion p
+  let istotal     = maybe False batotal     $ pbalanceassertion p
   actualbal' <-
     if isinclusive
     then
@@ -571,38 +616,40 @@ checkBalanceAssertionOneCommodityB p@Posting{paccount=assertedacct} assertedamt
       aquantity
         -- traceWith (("actual:"++).showAmountDebug)
         actualbalincomm
-
-    errmsg = printf (unlines
-                  [ "balance assertion: %s",
-                    "\nassertion details:",
-                    "date:       %s",
-                    "account:    %s%s",
-                    "commodity:  %s",
-                    -- "display precision:  %d",
-                    "calculated: %s", -- (at display precision: %s)",
-                    "asserted:   %s", -- (at display precision: %s)",
-                    "difference: %s"
-                  ])
-      (case ptransaction p of
-         Nothing -> "?" -- shouldn't happen
-         Just t ->  printf "%s\ntransaction:\n%s"
-                      (showSourcePos pos)
-                      (textChomp $ showTransaction t)
-                      :: String
-                      where
-                        pos = baposition $ fromJust $ pbalanceassertion p
-      )
-      (showDate $ postingDate p)
-      (T.unpack $ paccount p) -- XXX pack
-      (if isinclusive then " (and subs)" else "" :: String)
-      assertedcomm
+    errmsg = chomp $ printf (unlines
+      [ "%s:",
+        "%s\n",
+        "This balance assertion failed.",
+        -- "date:       %s",
+        "In account:    %s",
+        "and commodity: %s",
+        -- "display precision:  %d",
+        "this balance was asserted:     %s", -- (at display precision: %s)",
+        "but the calculated balance is: %s", -- (at display precision: %s)",
+        "a difference of:               %s",
+        "",
+        "Consider viewing this account's calculated balances to troubleshoot. Eg:",
+        "",
+        "hledger reg '%s'%s -I  # -f FILE"
+      ])
+      (sourcePosPretty pos)
+      (textChomp ex)
+      -- (showDate $ postingDate p)
+      (if isinclusive then printf "%-30s  (including subaccounts)" acct else acct)
+      (if istotal     then printf "%-30s  (no other commodities allowed)" (T.unpack assertedcomm) else (T.unpack assertedcomm))
       -- (asprecision $ astyle actualbalincommodity)  -- should be the standard display precision I think
-      (show $ aquantity actualbalincomm)
-      -- (showAmount actualbalincommodity)
       (show $ aquantity assertedamt)
       -- (showAmount assertedamt)
+      (show $ aquantity actualbalincomm)
+      -- (showAmount actualbalincommodity)
       (show $ aquantity assertedamt - aquantity actualbalincomm)
-
+      (acct ++ if isinclusive then "" else "$")
+      (if istotal then "" else (" cur:" ++ quoteForCommandLine (T.unpack assertedcomm)))
+      where
+        acct = T.unpack $ paccount p
+        ass = fromJust $ pbalanceassertion p  -- PARTIAL: fromJust won't fail, there is a balance assertion
+        pos = baposition ass
+        (_,_,_,ex) = makeBalanceAssertionErrorExcerpt p
   unless pass $ throwError errmsg
 
 -- | Throw an error if this posting is trying to do an illegal balance assignment.
@@ -619,7 +666,7 @@ checkBalanceAssignmentPostingDateB :: Posting -> Balancing s ()
 checkBalanceAssignmentPostingDateB p =
   when (hasBalanceAssignment p && isJust (pdate p)) $
     throwError $ chomp $ unlines [
-       "can't use balance assignment with custom posting date"
+       "Balance assignments and custom posting dates may not be combined."
       ,""
       ,chomp1 $ T.unpack $ maybe (T.unlines $ showPostingLines p) showTransaction $ ptransaction p
       ,"Balance assignments may not be used on postings with a custom posting date"
@@ -635,7 +682,7 @@ checkBalanceAssignmentUnassignableAccountB p = do
   unassignable <- R.asks bsUnassignable
   when (hasBalanceAssignment p && paccount p `S.member` unassignable) $
     throwError $ chomp $ unlines [
-       "can't use balance assignment with auto postings"
+       "Balance assignments and auto postings may not be combined."
       ,""
       ,chomp1 $ T.unpack $ maybe (T.unlines $ showPostingLines p) (showTransaction) $ ptransaction p
       ,"Balance assignments may not be used on accounts affected by auto posting rules"
@@ -653,11 +700,11 @@ tests_Balancing :: TestTree
 tests_Balancing =
   testGroup "Balancing" [
 
-      testCase "inferBalancingAmount" $ do
-         (fst <$> inferBalancingAmount M.empty nulltransaction) @?= Right nulltransaction
-         (fst <$> inferBalancingAmount M.empty nulltransaction{tpostings = ["a" `post` usd (-5), "b" `post` missingamt]}) @?=
+      testCase "transactionInferBalancingAmount" $ do
+         (fst <$> transactionInferBalancingAmount M.empty nulltransaction) @?= Right nulltransaction
+         (fst <$> transactionInferBalancingAmount M.empty nulltransaction{tpostings = ["a" `post` usd (-5), "b" `post` missingamt]}) @?=
            Right nulltransaction{tpostings = ["a" `post` usd (-5), "b" `post` usd 5]}
-         (fst <$> inferBalancingAmount M.empty nulltransaction{tpostings = ["a" `post` usd (-5), "b" `post` (eur 3 @@ usd 4), "c" `post` missingamt]}) @?=
+         (fst <$> transactionInferBalancingAmount M.empty nulltransaction{tpostings = ["a" `post` usd (-5), "b" `post` (eur 3 @@ usd 4), "c" `post` missingamt]}) @?=
            Right nulltransaction{tpostings = ["a" `post` usd (-5), "b" `post` (eur 3 @@ usd 4), "c" `post` usd 1]}
 
     , testGroup "balanceTransaction" [
