@@ -3,6 +3,7 @@
 {-# LANGUAGE Rank2Types          #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TupleSections       #-}
+{-# LANGUAGE RecordWildCards #-}
 
 {-|
 
@@ -26,8 +27,11 @@ module Hledger.Data.Journal (
   journalCommodityStyles,
   journalToCost,
   journalAddInferredEquityPostings,
+  journalInferCostsFromEquity,
+  journalMarkRedundantCosts,
   journalReverse,
   journalSetLastReadTime,
+  journalRenumberAccountDeclarations,
   journalPivot,
   -- * Filtering
   filterJournalTransactions,
@@ -36,6 +40,7 @@ module Hledger.Data.Journal (
   filterJournalAmounts,
   filterTransactionAmounts,
   filterTransactionPostings,
+  filterTransactionPostingsExtra,
   filterTransactionRelatedPostings,
   filterPostingAmount,
   -- * Mapping
@@ -62,6 +67,9 @@ module Hledger.Data.Journal (
   journalPayeesDeclared,
   journalPayeesUsed,
   journalPayeesDeclaredOrUsed,
+  journalTagsDeclared,
+  journalTagsUsed,
+  journalTagsDeclaredOrUsed,
   journalCommoditiesDeclared,
   journalCommodities,
   journalDateSpan,
@@ -83,24 +91,17 @@ module Hledger.Data.Journal (
   journalAddAccountTypes,
   journalPostingsAddAccountTags,
   -- journalPrices,
-  -- * Standard account types
-  journalBalanceSheetAccountQuery,
-  journalProfitAndLossAccountQuery,
-  journalRevenueAccountQuery,
-  journalExpenseAccountQuery,
-  journalAssetAccountQuery,
-  journalLiabilityAccountQuery,
-  journalEquityAccountQuery,
-  journalCashAccountQuery,
   journalConversionAccount,
   -- * Misc
   canonicalStyleFrom,
   nulljournal,
+  journalConcat,
   journalNumberTransactions,
   journalNumberAndTieTransactions,
   journalUntieTransactions,
   journalModifyTransactions,
   journalApplyAliases,
+  dbgJournalAcctDeclOrder,
   -- * Tests
   samplejournal,
   samplejournalMaybeExplicit,
@@ -115,15 +116,15 @@ import Control.Monad.State.Strict (StateT)
 import Data.Char (toUpper, isDigit)
 import Data.Default (Default(..))
 import Data.Foldable (toList)
-import Data.List ((\\), find, foldl', sortBy, union)
+import Data.List ((\\), find, foldl', sortBy, union, intercalate)
 import Data.List.Extra (nubSort)
 import qualified Data.Map.Strict as M
 import Data.Maybe (catMaybes, fromMaybe, mapMaybe, maybeToList)
 import qualified Data.Set as S
 import Data.Text (Text)
 import qualified Data.Text as T
-import Safe (headMay, headDef, maximumMay, minimumMay)
-import Data.Time.Calendar (Day, addDays, fromGregorian)
+import Safe (headMay, headDef, maximumMay, minimumMay, lastDef)
+import Data.Time.Calendar (Day, addDays, fromGregorian, diffDays)
 import Data.Time.Clock.POSIX (POSIXTime)
 import Data.Tree (Tree(..), flatten)
 import Text.Printf (printf)
@@ -139,16 +140,20 @@ import Hledger.Data.Transaction
 import Hledger.Data.TransactionModifier
 import Hledger.Data.Valuation
 import Hledger.Query
+import System.FilePath (takeFileName)
+import Data.Ord (comparing)
+import Hledger.Data.Dates (nulldate)
+import Data.List (sort)
 
 
 -- | A parser of text that runs in some monad, keeping a Journal as state.
-type JournalParser m a = StateT Journal (ParsecT CustomErr Text m) a
+type JournalParser m a = StateT Journal (ParsecT HledgerParseErrorData Text m) a
 
 -- | A parser of text that runs in some monad, keeping a Journal as
 -- state, that can throw an exception to end parsing, preventing
 -- further parser backtracking.
 type ErroringJournalParser m a =
-  StateT Journal (ParsecT CustomErr Text (ExceptT FinalParseError m)) a
+  StateT Journal (ParsecT HledgerParseErrorData Text (ExceptT FinalParseError m)) a
 
 -- deriving instance Show Journal
 instance Show Journal where
@@ -186,9 +191,7 @@ instance Show Journal where
 -- The semigroup instance for Journal is useful for two situations.
 --
 -- 1. concatenating finalised journals, eg with multiple -f options:
--- FIRST <> SECOND. The second's list fields are appended to the
--- first's, map fields are combined, transaction counts are summed,
--- the parse state of the second is kept.
+-- FIRST <> SECOND.
 --
 -- 2. merging a child parsed journal, eg with the include directive:
 -- CHILD <> PARENT. A parsed journal's data is in reverse order, so
@@ -196,8 +199,22 @@ instance Show Journal where
 --
 -- Note that (<>) is right-biased, so nulljournal is only a left identity.
 -- In particular, this prevents Journal from being a monoid.
-instance Semigroup Journal where
-  j1 <> j2 = Journal {
+instance Semigroup Journal where j1 <> j2 = j1 `journalConcat` j2
+
+-- | Merge two journals into one.
+-- Transaction counts are summed, map fields are combined,
+-- the second's list fields are appended to the first's,
+-- the second's parse state is kept.
+journalConcat :: Journal -> Journal -> Journal
+journalConcat j1 j2 =
+  let
+    f1 = takeFileName $ journalFilePath j1
+    f2 = maybe "(unknown)" takeFileName $ headMay $ jincludefilestack j2  -- XXX more accurate than journalFilePath for some reason
+  in
+    dbgJournalAcctDeclOrder ("journalConcat: " <> f1 <> " <> " <> f2 <> ", acct decls renumbered: ") $
+    journalRenumberAccountDeclarations $
+    dbgJournalAcctDeclOrder ("journalConcat: " <> f1 <> " <> " <> f2 <> ", acct decls           : ") $
+    Journal {
      jparsedefaultyear          = jparsedefaultyear          j2
     ,jparsedefaultcommodity     = jparsedefaultcommodity     j2
     ,jparsedecimalmark          = jparsedecimalmark          j2
@@ -207,6 +224,7 @@ instance Semigroup Journal where
     ,jparsetimeclockentries     = jparsetimeclockentries     j1 <> jparsetimeclockentries     j2
     ,jincludefilestack          = jincludefilestack j2
     ,jdeclaredpayees            = jdeclaredpayees            j1 <> jdeclaredpayees            j2
+    ,jdeclaredtags              = jdeclaredtags              j1 <> jdeclaredtags              j2
     ,jdeclaredaccounts          = jdeclaredaccounts          j1 <> jdeclaredaccounts          j2
     ,jdeclaredaccounttags       = jdeclaredaccounttags       j1 <> jdeclaredaccounttags       j2
     ,jdeclaredaccounttypes      = jdeclaredaccounttypes      j1 <> jdeclaredaccounttypes      j2
@@ -224,6 +242,33 @@ instance Semigroup Journal where
     ,jlastreadtime              = max (jlastreadtime j1) (jlastreadtime j2)
     }
 
+-- | Renumber all the account declarations. This is useful to call when
+-- finalising or concatenating Journals, to give account declarations
+-- a total order across files.
+journalRenumberAccountDeclarations :: Journal -> Journal
+journalRenumberAccountDeclarations j = j{jdeclaredaccounts=jdas'}
+  where
+    jdas' = [(a, adi{adideclarationorder=n}) | (n, (a,adi)) <- zip [1..] $ jdeclaredaccounts j]
+    -- the per-file declaration order saved during parsing is discarded,
+    -- it seems unneeded except perhaps for debugging
+
+-- | Debug log the ordering of a journal's account declarations
+-- (at debug level 5+).
+dbgJournalAcctDeclOrder :: String -> Journal -> Journal
+dbgJournalAcctDeclOrder prefix =
+  traceOrLogAtWith 5 ((prefix++) . showAcctDeclsSummary . jdeclaredaccounts)
+  where
+    showAcctDeclsSummary :: [(AccountName,AccountDeclarationInfo)] -> String
+    showAcctDeclsSummary adis
+      | length adis < (2*n+2) = "[" <> showadis adis <> "]"
+      | otherwise =
+          "[" <> showadis (take n adis) <> " ... " <> showadis (takelast n adis) <> "]"
+      where
+        n = 3
+        showadis = intercalate ", " . map showadi
+        showadi (a,adi) = "("<>show (adideclarationorder adi)<>","<>T.unpack a<>")"
+        takelast n' = reverse . take n' . reverse
+
 instance Default Journal where
   def = nulljournal
 
@@ -238,6 +283,7 @@ nulljournal = Journal {
   ,jparsetimeclockentries     = []
   ,jincludefilestack          = []
   ,jdeclaredpayees            = []
+  ,jdeclaredtags              = []
   ,jdeclaredaccounts          = []
   ,jdeclaredaccounttags       = M.empty
   ,jdeclaredaccounttypes      = M.empty
@@ -262,7 +308,7 @@ journalFilePaths :: Journal -> [FilePath]
 journalFilePaths = map fst . jfiles
 
 mainfile :: Journal -> (FilePath, Text)
-mainfile = headDef ("", "") . jfiles
+mainfile = headDef ("(unknown)", "") . jfiles
 
 addTransaction :: Transaction -> Journal -> Journal
 addTransaction t j = j { jtxns = t : jtxns j }
@@ -319,6 +365,20 @@ journalPayeesDeclaredOrUsed :: Journal -> [Payee]
 journalPayeesDeclaredOrUsed j = toList $ foldMap S.fromList
     [journalPayeesDeclared j, journalPayeesUsed j]
 
+-- | Sorted unique tag names declared by tag directives in this journal.
+journalTagsDeclared :: Journal -> [TagName]
+journalTagsDeclared = nubSort . map fst . jdeclaredtags
+
+-- | Sorted unique tag names used in this journal (in account directives, transactions, postings..)
+journalTagsUsed :: Journal -> [TagName]
+journalTagsUsed j = nubSort $ map fst $ concatMap transactionAllTags $ jtxns j
+  -- tags used in all transactions and postings and postings' accounts
+
+-- | Sorted unique tag names used in transactions or declared by tag directives in this journal.
+journalTagsDeclaredOrUsed :: Journal -> [TagName]
+journalTagsDeclaredOrUsed j = toList $ foldMap S.fromList
+    [journalTagsDeclared j, journalTagsUsed j]
+
 -- | Sorted unique account names posted to by this journal's transactions.
 journalAccountNamesUsed :: Journal -> [AccountName]
 journalAccountNamesUsed = accountNamesFromPostings . journalPostings
@@ -368,25 +428,43 @@ journalAccountTags Journal{jdeclaredaccounttags} a = M.findWithDefault [] a jdec
 -- | Which tags are in effect for this account, including tags inherited from parent accounts ?
 journalInheritedAccountTags :: Journal -> AccountName -> [Tag]
 journalInheritedAccountTags j a =
-  foldl' (\ts a -> ts `union` journalAccountTags j a) [] as
+  foldl' (\ts a' -> ts `union` journalAccountTags j a') [] as
   where
     as = a : parentAccountNames a
 -- PERF: cache in journal ?
 
+type DateWeightedSimilarityScore = Double
+type SimilarityScore = Double
+type Age = Integer
+
 -- | Find up to N most similar and most recent transactions matching
--- the given transaction description and query. Transactions are
--- listed with their description's similarity score (see
--- compareDescriptions), sorted by highest score and then by date.
--- Only transactions with a similarity score greater than a minimum
--- threshold (currently 0) are returned.
-journalTransactionsSimilarTo :: Journal -> Query -> Text -> Int -> [(Double,Transaction)]
-journalTransactionsSimilarTo Journal{jtxns} q desc n =
+-- the given transaction description and query and exceeding the given
+-- description similarity score (0 to 1, see compareDescriptions).
+-- Returns transactions along with
+-- their age in days compared to the latest transaction date,
+-- their description similarity score,
+-- and a heuristically date-weighted variant of this that favours more recent transactions.
+journalTransactionsSimilarTo :: Journal -> Text -> Query -> SimilarityScore -> Int
+  -> [(DateWeightedSimilarityScore, Age, SimilarityScore, Transaction)]
+journalTransactionsSimilarTo Journal{jtxns} desc q similaritythreshold n =
   take n $
-  sortBy (\(s1,t1) (s2,t2) -> compare (s2,tdate t2) (s1,tdate t1)) $
-  filter ((> threshold).fst)
+  dbg1With (
+    unlines . 
+    ("up to 30 transactions above description similarity threshold "<>show similaritythreshold<>" ordered by recency-weighted similarity:":) .
+    take 30 .
+    map ( \(w,a,s,Transaction{..}) -> printf "weighted:%8.3f  age:%4d similarity:%5.3f  %s %s" w a s (show tdate) tdescription )) $
+  sortBy (comparing (negate.first4)) $
+  map (\(s,t) -> (weightedScore (s,t), age t, s, t)) $
+  filter ((> similaritythreshold).fst)
   [(compareDescriptions desc $ tdescription t, t) | t <- jtxns, q `matchesTransaction` t]
   where
-    threshold = 0
+    latest = lastDef nulldate $ sort $ map tdate jtxns
+    age = diffDays latest . tdate
+    -- Combine similarity and recency heuristically. This gave decent results
+    -- in my "find most recent invoice" use case in 2023-03,
+    -- but will probably need more attention.
+    weightedScore :: (Double, Transaction) -> Double
+    weightedScore (s, t) = 100 * s - fromIntegral (age t) / 4
 
 -- | Return a similarity score from 0 to 1.5 for two transaction descriptions. 
 -- This is based on compareStrings, with the following modifications:
@@ -428,109 +506,6 @@ letterPairs :: String -> [String]
 letterPairs (a:b:rest) = [a,b] : letterPairs (b:rest)
 letterPairs _ = []
 
--- queries for standard account types
-
--- | Get a query for accounts of the specified types in this journal. 
--- Account types include:
--- Asset, Liability, Equity, Revenue, Expense, Cash, Conversion.
--- For each type, if no accounts were declared with this type, the query 
--- will instead match accounts with names matched by the case-insensitive 
--- regular expression provided as a fallback.
--- The query will match all accounts which were declared as one of
--- these types (by account directives with the type: tag), plus all their 
--- subaccounts which have not been declared as some other type.
---
--- This is older code pre-dating 2022's expansion of account types.
-journalAccountTypeQuery :: [AccountType] -> Regexp -> Journal -> Query
-journalAccountTypeQuery atypes fallbackregex Journal{jdeclaredaccounttypes} =
-  let
-    declaredacctsoftype :: [AccountName] =
-      concat $ mapMaybe (`M.lookup` jdeclaredaccounttypes) atypes
-  in case declaredacctsoftype of
-    [] -> Acct fallbackregex
-    as -> And $ Or acctnameRegexes : if null differentlyTypedRegexes then [] else [ Not $ Or differentlyTypedRegexes ]
-      where
-        -- XXX Query isn't able to match account type since that requires extra info from the journal.
-        -- So we do a hacky search by name instead.
-        acctnameRegexes = map (Acct . accountNameToAccountRegex) as
-        differentlyTypedRegexes = map (Acct . accountNameToAccountRegex) differentlytypedsubs
-
-        differentlytypedsubs = concat
-          [subs | (t,bs) <- M.toList jdeclaredaccounttypes
-              , t `notElem` atypes
-              , let subs = [b | b <- bs, any (`isAccountNamePrefixOf` b) as]
-          ]
-
--- | A query for accounts in this journal which have been
--- declared as Asset (or Cash, a subtype of Asset) by account directives, 
--- or otherwise for accounts with names matched by the case-insensitive 
--- regular expression @^assets?(:|$)@.
-journalAssetAccountQuery :: Journal -> Query
-journalAssetAccountQuery j =
-  Or [
-     journalAccountTypeQuery [Asset] assetAccountRegex j
-    ,journalCashAccountOnlyQuery j
-  ]
-
--- | A query for Cash (liquid asset) accounts in this journal, ie accounts
--- declared as Cash by account directives, or otherwise accounts whose
--- names match the case-insensitive regular expression
--- @(^assets:(.+:)?(cash|bank)(:|$)@.
-journalCashAccountQuery :: Journal -> Query
-journalCashAccountQuery = journalAccountTypeQuery [Cash] cashAccountRegex
-
--- | A query for accounts in this journal specifically declared as Cash by 
--- account directives, or otherwise the None query.
-journalCashAccountOnlyQuery :: Journal -> Query
-journalCashAccountOnlyQuery j
-  -- Cash accounts are declared; get a query for them (the fallback regex won't be used)
-  | Cash `M.member` jdeclaredaccounttypes j = journalAccountTypeQuery [Cash] notused j
-  | otherwise = None
-  where notused = error' "journalCashAccountOnlyQuery: this should not have happened!"  -- PARTIAL:
-
--- | A query for accounts in this journal which have been
--- declared as Liability by account directives, or otherwise for
--- accounts with names matched by the case-insensitive regular expression
--- @^(debts?|liabilit(y|ies))(:|$)@.
-journalLiabilityAccountQuery :: Journal -> Query
-journalLiabilityAccountQuery = journalAccountTypeQuery [Liability] liabilityAccountRegex
-
--- | A query for accounts in this journal which have been
--- declared as Equity by account directives, or otherwise for
--- accounts with names matched by the case-insensitive regular expression
--- @^equity(:|$)@.
-journalEquityAccountQuery :: Journal -> Query
-journalEquityAccountQuery = journalAccountTypeQuery [Equity] equityAccountRegex
-
--- | A query for accounts in this journal which have been
--- declared as Revenue by account directives, or otherwise for
--- accounts with names matched by the case-insensitive regular expression
--- @^(income|revenue)s?(:|$)@.
-journalRevenueAccountQuery :: Journal -> Query
-journalRevenueAccountQuery = journalAccountTypeQuery [Revenue] revenueAccountRegex
-
--- | A query for accounts in this journal which have been
--- declared as Expense by account directives, or otherwise for
--- accounts with names matched by the case-insensitive regular expression
--- @^expenses?(:|$)@.
-journalExpenseAccountQuery  :: Journal -> Query
-journalExpenseAccountQuery = journalAccountTypeQuery [Expense] expenseAccountRegex
-
--- | A query for Asset, Liability & Equity accounts in this journal.
--- Cf <http://en.wikipedia.org/wiki/Chart_of_accounts#Balance_Sheet_Accounts>.
-journalBalanceSheetAccountQuery :: Journal -> Query
-journalBalanceSheetAccountQuery j = Or [journalAssetAccountQuery j
-                                       ,journalLiabilityAccountQuery j
-                                       ,journalEquityAccountQuery j
-                                       ]
-
--- | A query for Profit & Loss accounts in this journal.
--- Cf <http://en.wikipedia.org/wiki/Chart_of_accounts#Profit_.26_Loss_accounts>.
-journalProfitAndLossAccountQuery  :: Journal -> Query
-journalProfitAndLossAccountQuery j = Or [journalRevenueAccountQuery j
-                                        ,journalExpenseAccountQuery j
-                                        ]
-
 -- | The 'AccountName' to use for automatically generated conversion postings.
 journalConversionAccount :: Journal -> AccountName
 journalConversionAccount =
@@ -538,7 +513,7 @@ journalConversionAccount =
     . M.findWithDefault [] Conversion
     . jdeclaredaccounttypes
 
--- Newer account type functionality.
+-- Newer account type code.
 
 journalAccountType :: Journal -> AccountName -> Maybe AccountType
 journalAccountType Journal{jaccounttypes} = accountNameType jaccounttypes
@@ -553,19 +528,21 @@ journalAccountTypes :: Journal -> M.Map AccountName AccountType
 journalAccountTypes j = M.fromList [(a,acctType) | (a, Just (acctType,_)) <- flatten t']
   where
     t = accountNameTreeFrom $ journalAccountNames j :: Tree AccountName
-    t' = settypes Nothing t :: Tree (AccountName, Maybe (AccountType, Bool))
     -- Map from the top of the account tree down to the leaves, propagating
     -- account types downward. Keep track of whether the account is declared
     -- (True), in which case the parent account should be preferred, or merely
     -- inferred (False), in which case the inferred type should be preferred.
-    settypes :: Maybe (AccountType, Bool) -> Tree AccountName -> Tree (AccountName, Maybe (AccountType, Bool))
-    settypes mparenttype (Node a subs) = Node (a, mtype) (map (settypes mtype) subs)
+    t' = settypes Nothing t :: Tree (AccountName, Maybe (AccountType, Bool))
       where
-        mtype = M.lookup a declaredtypes <|> minferred
-        minferred = if maybe False snd mparenttype
-                       then mparenttype
-                       else (,False) <$> accountNameInferType a <|> mparenttype
-    declaredtypes = (,True) <$> journalDeclaredAccountTypes j
+        settypes :: Maybe (AccountType, Bool) -> Tree AccountName -> Tree (AccountName, Maybe (AccountType, Bool))
+        settypes mparenttype (Node a subs) = Node (a, mtype) (map (settypes mtype) subs)
+          where
+            mtype = M.lookup a declaredtypes <|> minferred
+              where 
+                declaredtypes = (,True) <$> journalDeclaredAccountTypes j
+                minferred = if maybe False snd mparenttype
+                            then mparenttype
+                            else (,False) <$> accountNameInferType a <|> mparenttype
 
 -- | Build a map of the account types explicitly declared.
 journalDeclaredAccountTypes :: Journal -> M.Map AccountName AccountType
@@ -808,13 +785,12 @@ journalUntieTransactions t@Transaction{tpostings=ps} = t{tpostings=map (\p -> p{
 -- postings to transactions, eg). Or if a modifier rule fails to parse,
 -- return the error message. A reference date is provided to help interpret
 -- relative dates in transaction modifier queries.
-journalModifyTransactions :: Day -> Journal -> Either String Journal
-journalModifyTransactions d j =
-    case modifyTransactions (journalAccountType j) (journalInheritedAccountTags j) (journalCommodityStyles j) d (jtxnmodifiers j) (jtxns j) of
-      Right ts -> Right j{jtxns=ts}
-      Left err -> Left err
-
---
+-- The first argument selects whether to add visible tags to generated postings & modified transactions.
+journalModifyTransactions :: Bool -> Day -> Journal -> Either String Journal
+journalModifyTransactions verbosetags d j =
+  case modifyTransactions (journalAccountType j) (journalInheritedAccountTags j) (journalCommodityStyles j) d verbosetags (jtxnmodifiers j) (jtxns j) of
+    Right ts -> Right j{jtxns=ts}
+    Left err -> Left err
 
 -- | Choose and apply a consistent display style to the posting
 -- amounts in each commodity (see journalCommodityStyles).
@@ -926,26 +902,12 @@ canonicalStyle a b = a{asprecision=prec, asdecimalpoint=decmark, asdigitgroups=m
 journalInferMarketPricesFromTransactions :: Journal -> Journal
 journalInferMarketPricesFromTransactions j =
   j{jinferredmarketprices =
-       dbg4 "jinferredmarketprices" $
-       mapMaybe postingInferredmarketPrice $ journalPostings j
+       dbg4 "jinferredmarketprices" .
+       map priceDirectiveToMarketPrice .
+       concatMap postingPriceDirectivesFromCost $
+       journalPostings j
    }
 
--- | Make a market price equivalent to this posting's amount's unit
--- price, if any. If the posting amount is multicommodity, only the
--- first commodity amount is considered.
-postingInferredmarketPrice :: Posting -> Maybe MarketPrice
-postingInferredmarketPrice p@Posting{pamount} =
-    -- convert any total prices to unit prices
-    case amountsRaw $ mixedAmountTotalPriceToUnitPrice pamount of
-      Amount{acommodity=fromcomm, aprice = Just (UnitPrice Amount{acommodity=tocomm, aquantity=rate})}:_ ->
-        Just MarketPrice {
-           mpdate = postingDate p
-          ,mpfrom = fromcomm
-          ,mpto   = tocomm
-          ,mprate = rate
-          }
-      _ -> Nothing
-
 -- | Convert all this journal's amounts to cost using the transaction prices, if any.
 -- The journal's commodity styles are applied to the resulting amounts.
 journalToCost :: ConversionOp -> Journal -> Journal
@@ -954,11 +916,28 @@ journalToCost cost j@Journal{jtxns=ts} = j{jtxns=map (transactionToCost styles c
     styles = journalCommodityStyles j
 
 -- | Add inferred equity postings to a 'Journal' using transaction prices.
-journalAddInferredEquityPostings :: Journal -> Journal
-journalAddInferredEquityPostings j = journalMapTransactions (transactionAddInferredEquityPostings equityAcct) j
+journalAddInferredEquityPostings :: Bool -> Journal -> Journal
+journalAddInferredEquityPostings verbosetags j = journalMapTransactions (transactionAddInferredEquityPostings verbosetags equityAcct) j
   where
     equityAcct = journalConversionAccount j
 
+-- | Add costs inferred from equity conversion postings, where needed and possible.
+-- See hledger manual > Inferring cost from equity postings.
+journalInferCostsFromEquity :: Journal -> Either String Journal
+journalInferCostsFromEquity j = do
+    ts <- mapM (transactionInferCostsFromEquity False $ jaccounttypes j) $ jtxns j
+    return j{jtxns=ts}
+
+-- | Do just the internal tagging that is normally done by journalInferCostsFromEquity,
+-- identifying equity conversion postings and, in particular, postings which have redundant costs.
+-- Tagging the latter is useful as it allows them to be ignored during transaction balancedness checking.
+-- And that allows journalInferCostsFromEquity to be postponed till after transaction balancing,
+-- when it will have more information (amounts) to work with.
+journalMarkRedundantCosts :: Journal -> Either String Journal
+journalMarkRedundantCosts j = do
+    ts <- mapM (transactionInferCostsFromEquity True $ jaccounttypes j) $ jtxns j
+    return j{jtxns=ts}
+
 -- -- | Get this journal's unique, display-preference-canonicalised commodities, by symbol.
 -- journalCanonicalCommodities :: Journal -> M.Map String CommoditySymbol
 -- journalCanonicalCommodities j = canonicaliseCommodities $ journalAmountCommodities j
@@ -1005,17 +984,17 @@ journalStyleInfluencingAmounts j =
 
 -- overcomplicated/unused amount traversal stuff
 --
--- | Get an ordered list of 'AmountStyle's from the amounts in this
+--  Get an ordered list of 'AmountStyle's from the amounts in this
 -- journal which influence canonical amount display styles. See
 -- traverseJournalAmounts.
 -- journalAmounts :: Journal -> [Amount]
 -- journalAmounts = getConst . traverseJournalAmounts (Const . (:[]))
 --
--- | Apply a transformation to the journal amounts traversed by traverseJournalAmounts.
+--  Apply a transformation to the journal amounts traversed by traverseJournalAmounts.
 -- overJournalAmounts :: (Amount -> Amount) -> Journal -> Journal
 -- overJournalAmounts f = runIdentity . traverseJournalAmounts (Identity . f)
 --
--- | A helper that traverses over most amounts in the journal,
+--  A helper that traverses over most amounts in the journal,
 -- in particular the ones which influence canonical amount display styles,
 -- processing them with the given applicative function.
 --
@@ -1044,7 +1023,7 @@ journalStyleInfluencingAmounts j =
 --     pamt  g p          = (\amt -> p {pamount  =amt}) <$> g (pamount p)
 --     amts  g (Mixed as) = Mixed <$> g as
 
--- | The fully specified date span enclosing the dates (primary or secondary)
+-- | The fully specified exact date span enclosing the dates (primary or secondary)
 -- of all this journal's transactions and postings, or DateSpan Nothing Nothing
 -- if there are none.
 journalDateSpan :: Bool -> Journal -> DateSpan
@@ -1061,7 +1040,7 @@ journalDateSpanBothDates = journalDateSpanHelper Nothing
 -- uses both primary and secondary dates.
 journalDateSpanHelper :: Maybe WhichDate -> Journal -> DateSpan
 journalDateSpanHelper whichdate j =
-    DateSpan (minimumMay dates) (addDays 1 <$> maximumMay dates)
+    DateSpan (Exact <$> minimumMay dates) (Exact . addDays 1 <$> maximumMay dates)
   where
     dates    = pdates ++ tdates
     tdates   = concatMap gettdate ts
@@ -1079,12 +1058,12 @@ journalDateSpanHelper whichdate j =
 -- | The earliest of this journal's transaction and posting dates, or
 -- Nothing if there are none.
 journalStartDate :: Bool -> Journal -> Maybe Day
-journalStartDate secondary j = b where DateSpan b _ = journalDateSpan secondary j
+journalStartDate secondary j = fromEFDay <$> b where DateSpan b _ = journalDateSpan secondary j
 
 -- | The "exclusive end date" of this journal: the day following its latest transaction 
 -- or posting date, or Nothing if there are none.
 journalEndDate :: Bool -> Journal -> Maybe Day
-journalEndDate secondary j = e where DateSpan _ e = journalDateSpan secondary j
+journalEndDate secondary j = fromEFDay <$> e where DateSpan _ e = journalDateSpan secondary j
 
 -- | The latest of this journal's transaction and posting dates, or
 -- Nothing if there are none.
@@ -1111,6 +1090,7 @@ postingPivot fieldortagname p = p{paccount = pivotedacct, poriginal = Just $ ori
       | Just t <- ptransaction p, fieldortagname == "description" = tdescription t
       | Just t <- ptransaction p, fieldortagname == "payee"       = transactionPayee t
       | Just t <- ptransaction p, fieldortagname == "note"        = transactionNote t
+      | Just t <- ptransaction p, fieldortagname == "status"      = T.pack . show . tstatus $ t
       | Just (_, value) <- postingFindTag fieldortagname p        = value
       | otherwise                                                 = ""
 
@@ -1295,25 +1275,5 @@ tests_Journal = testGroup "Journal" [
                               }
               ]
       }
-    @?= (DateSpan (Just $ fromGregorian 2014 1 10) (Just $ fromGregorian 2014 10 11))
-
-  ,testGroup "standard account type queries" $
-    let
-      j = samplejournal
-      journalAccountNamesMatching :: Query -> Journal -> [AccountName]
-      journalAccountNamesMatching q = filter (q `matchesAccount`) . journalAccountNames
-      namesfrom qfunc = journalAccountNamesMatching (qfunc j) j
-    in [testCase "assets"      $ assertEqual "" ["assets","assets:bank","assets:bank:checking","assets:bank:saving","assets:cash"]
-         (namesfrom journalAssetAccountQuery)
-       ,testCase "cash"        $ assertEqual "" ["assets:bank","assets:bank:checking","assets:bank:saving","assets:cash"]
-         (namesfrom journalCashAccountQuery)
-       ,testCase "liabilities" $ assertEqual "" ["liabilities","liabilities:debts"]
-         (namesfrom journalLiabilityAccountQuery)
-       ,testCase "equity"      $ assertEqual "" []
-         (namesfrom journalEquityAccountQuery)
-       ,testCase "income"      $ assertEqual "" ["income","income:gifts","income:salary"]
-         (namesfrom journalRevenueAccountQuery)
-       ,testCase "expenses"    $ assertEqual "" ["expenses","expenses:food","expenses:supplies"]
-         (namesfrom journalExpenseAccountQuery)
-       ]
+    @?= (DateSpan (Just $ Exact $ fromGregorian 2014 1 10) (Just $ Exact $ fromGregorian 2014 10 11))
   ]
