@@ -61,19 +61,22 @@ module Hledger.Data.Posting (
   postingsAsLines,
   showAccountName,
   renderCommentLines,
+  showBalanceAssertion,
   -- * misc.
   postingTransformAmount,
   postingApplyValuation,
   postingToCost,
   postingAddInferredEquityPostings,
+  postingPriceDirectivesFromCost,
   tests_Posting
 )
 where
 
 import Data.Default (def)
 import Data.Foldable (asum)
+import Data.Function ((&))
 import qualified Data.Map as M
-import Data.Maybe (fromMaybe, isJust)
+import Data.Maybe (fromMaybe, isJust, mapMaybe)
 import Data.List (foldl', sort, union)
 import qualified Data.Set as S
 import Data.Text (Text)
@@ -84,7 +87,7 @@ import Data.Time.Calendar (Day)
 import Safe (maximumBound)
 import Text.DocLayout (realLength)
 
-import Text.Tabular.AsciiWide
+import Text.Tabular.AsciiWide hiding (render)
 
 import Hledger.Utils
 import Hledger.Data.Types
@@ -191,6 +194,7 @@ showPostingLines p = first3 $ postingAsLines False False maxacctwidth maxamtwidt
 -- if onelineamounts is true, these amounts are shown on one line,
 -- comma-separated, and the output will not be valid journal syntax.
 -- Otherwise, they are shown as several similar postings, one per commodity.
+-- When the posting has a balance assertion, it is attached to the last of these postings.
 --
 -- The output will appear to be a balanced transaction.
 -- Amounts' display precisions, which may have been limited by commodity
@@ -241,16 +245,11 @@ postingAsLines elideamount onelineamounts acctwidth amtwidth p =
                               , Cell BottomLeft [assertion]
                               , textCell BottomLeft samelinecomment
                               ]
-                    | amt <- shownAmounts]
+                    | (amt,assertion) <- shownAmountsAssertions]
     render = renderRow def{tableBorders=False, borderSpaces=False} . Group NoLine . map Header
     pad amt = WideBuilder (TB.fromText $ T.replicate w " ") w <> amt
       where w = max 12 amtwidth - wbWidth amt  -- min. 12 for backwards compatibility
 
-    assertion = maybe mempty ((WideBuilder (TB.singleton ' ') 1 <>).showBalanceAssertion) $ pbalanceassertion p
-    -- pad to the maximum account name width, plus 2 to leave room for status flags, to keep amounts aligned
-    statusandaccount = lineIndent . fitText (Just $ 2 + acctwidth) Nothing False True $ pstatusandacct p
-    thisacctwidth = realLength $ pacctstr p
-
     pacctstr p' = showAccountName Nothing (ptype p') (paccount p')
     pstatusandacct p' = pstatusprefix p' <> pacctstr p'
     pstatusprefix p' = case pstatus p' of
@@ -266,6 +265,17 @@ postingAsLines elideamount onelineamounts acctwidth amtwidth p =
       | otherwise   = showMixedAmountLinesB noColour{displayOneLine=onelineamounts} $ pamount p
     thisamtwidth = maximumBound 0 $ map wbWidth shownAmounts
 
+    -- when there is a balance assertion, show it only on the last posting line
+    shownAmountsAssertions = zip shownAmounts shownAssertions
+      where
+        shownAssertions = replicate (length shownAmounts - 1) mempty ++ [assertion]
+          where
+            assertion = maybe mempty ((WideBuilder (TB.singleton ' ') 1 <>).showBalanceAssertion) $ pbalanceassertion p
+
+    -- pad to the maximum account name width, plus 2 to leave room for status flags, to keep amounts aligned
+    statusandaccount = lineIndent . fitText (Just $ 2 + acctwidth) Nothing False True $ pstatusandacct p
+    thisacctwidth = realLength $ pacctstr p
+
     (samelinecomment, newlinecomments) =
       case renderCommentLines (pcomment p) of []   -> ("",[])
                                               c:cs -> (c,cs)
@@ -394,7 +404,7 @@ postingApplyAliases aliases p@Posting{paccount} =
     Right a -> Right p{paccount=a}
     Left e  -> Left err
       where
-        err = "problem while applying account aliases:\n" ++ pshow aliases 
+        err = "problem while applying account aliases:\n" ++ pshow aliases
           ++ "\n to account name: "++T.unpack paccount++"\n "++e
 
 -- | Choose and apply a consistent display style to the posting
@@ -418,19 +428,26 @@ postingApplyValuation priceoracle styles periodlast today v p =
 
 -- | Maybe convert this 'Posting's amount to cost, and apply apply appropriate
 -- amount styles.
-postingToCost :: M.Map CommoditySymbol AmountStyle -> ConversionOp -> Posting -> Posting
-postingToCost _      NoConversionOp p = p
-postingToCost styles ToCost         p = postingTransformAmount (styleMixedAmount styles . mixedAmountCost) p
+postingToCost :: M.Map CommoditySymbol AmountStyle -> ConversionOp -> Posting -> Maybe Posting
+postingToCost _      NoConversionOp p = Just p
+postingToCost styles ToCost         p
+    -- If this is a conversion posting with a matched transaction price posting, ignore it
+    | "_conversion-matched" `elem` map fst (ptags p) && noCost = Nothing
+    | otherwise = Just $ postingTransformAmount (styleMixedAmount styles . mixedAmountCost) p
+  where
+    noCost = (not . any (isJust . aprice) . amountsRaw) $ pamount p
 
 -- | Generate inferred equity postings from a 'Posting' using transaction prices.
-postingAddInferredEquityPostings :: Text -> Posting -> [Posting]
-postingAddInferredEquityPostings equityAcct p = taggedPosting : concatMap conversionPostings priceAmounts
+-- Make sure not to generate equity postings when there are already matched
+-- conversion postings.
+postingAddInferredEquityPostings :: Bool -> Text -> Posting -> [Posting]
+postingAddInferredEquityPostings verbosetags equityAcct p
+    | "_price-matched" `elem` map fst (ptags p) = [p]
+    | otherwise = taggedPosting : concatMap conversionPostings priceAmounts
   where
     taggedPosting
       | null priceAmounts = p
-      | otherwise         = p{ pcomment = pcomment p `commentAddTag` priceTag
-                             , ptags = priceTag : ptags p
-                             }
+      | otherwise         = p{ ptags = ("_price-matched","") : ptags p }
     conversionPostings amt = case aprice amt of
         Nothing -> []
         Just _  -> [ cp{ paccount = accountPrefix <> amtCommodity
@@ -444,8 +461,11 @@ postingAddInferredEquityPostings equityAcct p = taggedPosting : concatMap conver
         cost = amountCost amt
         amtCommodity  = commodity amt
         costCommodity = commodity cost
-        cp = p{ pcomment = pcomment p `commentAddTag` ("generated-posting","")
-              , ptags = [("generated-posting", ""), ("_generated-posting", "")]
+        cp = p{ pcomment = pcomment p & (if verbosetags then (`commentAddTag` ("generated-posting","conversion")) else id)
+              , ptags    =
+                   ("_conversion-matched","") : -- implementation-specific internal tag, not for users
+                   ("_generated-posting","conversion") :
+                   (if verbosetags then [("generated-posting", "conversion")] else [])
               , pbalanceassertion = Nothing
               , poriginal = Nothing
               }
@@ -453,9 +473,14 @@ postingAddInferredEquityPostings equityAcct p = taggedPosting : concatMap conver
         -- Take the commodity of an amount and collapse consecutive spaces to a single space
         commodity = T.unwords . filter (not . T.null) . T.words . acommodity
 
-    priceTag = ("cost", T.strip . wbToText $ foldMap showAmountPrice priceAmounts)
     priceAmounts = filter (isJust . aprice) . amountsRaw $ pamount p
 
+-- | Make a market price equivalent to this posting's amount's unit
+-- price, if any.
+postingPriceDirectivesFromCost :: Posting -> [PriceDirective]
+postingPriceDirectivesFromCost p@Posting{pamount} =
+    mapMaybe (amountPriceDirectiveFromCost $ postingDate p) $ amountsRaw pamount
+
 -- | Apply a transform function to this posting's amount.
 postingTransformAmount :: (MixedAmount -> MixedAmount) -> Posting -> Posting
 postingTransformAmount f p@Posting{pamount=a} = p{pamount=f a}
@@ -483,7 +508,7 @@ commentAddTag c (t,v)
 -- A space is inserted following the colon, before the value.
 commentAddTagNextLine :: Text -> Tag -> Text
 commentAddTagNextLine cmt (t,v) =
-  cmt <> (if "\n" `T.isSuffixOf` cmt then "" else "\n") <> t <> ": " <> v 
+  cmt <> (if "\n" `T.isSuffixOf` cmt then "" else "\n") <> t <> ": " <> v
 
 
 -- tests
