@@ -7,6 +7,7 @@ tags.
 
 -}
 
+{-# LANGUAGE MultiWayIf        #-}
 {-# LANGUAGE NamedFieldPuns    #-}
 {-# LANGUAGE OverloadedStrings #-}
 
@@ -27,9 +28,11 @@ module Hledger.Data.Transaction
 , transactionApplyValuation
 , transactionToCost
 , transactionAddInferredEquityPostings
+, transactionInferCostsFromEquity
 , transactionApplyAliases
 , transactionMapPostings
 , transactionMapPostingAmounts
+, partitionAndCheckConversionPostings
   -- nonzerobalanceerror
   -- * date operations
 , transactionDate2
@@ -41,18 +44,25 @@ module Hledger.Data.Transaction
   -- * rendering
 , showTransaction
 , showTransactionOneLineAmounts
+, showTransactionLineFirstPart
 , transactionFile
+  -- * transaction errors
+, annotateErrorWithTransaction
   -- * tests
 , tests_Transaction
 ) where
 
-import Data.Maybe (fromMaybe)
+import Control.Monad.Trans.State (StateT(..), evalStateT)
+import Data.Bifunctor (first)
+import Data.Foldable (foldrM)
+import Data.Maybe (fromMaybe, isJust, mapMaybe)
+import Data.Semigroup (Endo(..))
 import Data.Text (Text)
+import qualified Data.Map as M
 import qualified Data.Text as T
 import qualified Data.Text.Lazy as TL
 import qualified Data.Text.Lazy.Builder as TB
 import Data.Time.Calendar (Day, fromGregorian)
-import qualified Data.Map as M
 
 import Hledger.Utils
 import Hledger.Data.Types
@@ -137,18 +147,22 @@ showTransactionHelper onelineamounts t =
     <> foldMap ((<> newline) . TB.fromText) (postingsAsLines onelineamounts $ tpostings t)
     <> newline
   where
-    descriptionline = T.stripEnd $ T.concat [date, status, code, desc, samelinecomment]
-    date = showDate (tdate t) <> maybe "" (("="<>) . showDate) (tdate2 t)
-    status | tstatus t == Cleared = " *"
-           | tstatus t == Pending = " !"
-           | otherwise            = ""
-    code = if T.null (tcode t) then "" else wrap " (" ")" $ tcode t
+    descriptionline = T.stripEnd $ showTransactionLineFirstPart t <> T.concat [desc, samelinecomment]
     desc = if T.null d then "" else " " <> d where d = tdescription t
     (samelinecomment, newlinecomments) =
       case renderCommentLines (tcomment t) of []   -> ("",[])
                                               c:cs -> (c,cs)
     newline = TB.singleton '\n'
 
+-- Useful when rendering error messages.
+showTransactionLineFirstPart t = T.concat [date, status, code]
+  where
+    date = showDate (tdate t) <> maybe "" (("="<>) . showDate) (tdate2 t)
+    status | tstatus t == Cleared = " *"
+           | tstatus t == Pending = " !"
+           | otherwise            = ""
+    code = if T.null (tcode t) then "" else wrap " (" ")" $ tcode t
+
 hasRealPostings :: Transaction -> Bool
 hasRealPostings = not . null . realPostings
 
@@ -205,12 +219,149 @@ transactionApplyValuation priceoracle styles periodlast today v =
 -- | Maybe convert this 'Transaction's amounts to cost and apply the
 -- appropriate amount styles.
 transactionToCost :: M.Map CommoditySymbol AmountStyle -> ConversionOp -> Transaction -> Transaction
-transactionToCost styles cost = transactionMapPostings (postingToCost styles cost)
+transactionToCost styles cost t = t{tpostings = mapMaybe (postingToCost styles cost) $ tpostings t}
 
 -- | Add inferred equity postings to a 'Transaction' using transaction prices.
-transactionAddInferredEquityPostings :: AccountName -> Transaction -> Transaction
-transactionAddInferredEquityPostings equityAcct t =
-    t{tpostings=concatMap (postingAddInferredEquityPostings equityAcct) $ tpostings t}
+transactionAddInferredEquityPostings :: Bool -> AccountName -> Transaction -> Transaction
+transactionAddInferredEquityPostings verbosetags equityAcct t =
+    t{tpostings=concatMap (postingAddInferredEquityPostings verbosetags equityAcct) $ tpostings t}
+
+type IdxPosting = (Int, Posting)
+
+-- WARNING: twisty code ahead
+
+-- | Add costs inferred from equity postings in this transaction.
+-- For every adjacent pair of conversion postings, it will first search the postings
+-- with costs to see if any match. If so, it will tag these as matched.
+-- If no postings with costs match, it will then search the postings without costs,
+-- and will match the first such posting which matches one of the conversion amounts.
+-- If it finds a match, it will add a cost and then tag it.
+-- If the first argument is true, do a dry run instead: identify and tag
+-- the costful and conversion postings, but don't add costs.
+transactionInferCostsFromEquity :: Bool -> M.Map AccountName AccountType -> Transaction -> Either String Transaction
+transactionInferCostsFromEquity dryrun acctTypes t = first (annotateErrorWithTransaction t . T.unpack) $ do
+    (conversionPairs, stateps) <- partitionAndCheckConversionPostings False acctTypes npostings
+    f <- transformIndexedPostingsF (addCostsToPostings dryrun) conversionPairs stateps
+    return t{tpostings = map (snd . f) npostings}
+  where
+    -- Include indices for postings
+    npostings = zip [0..] $ tpostings t
+    transformIndexedPostingsF f = evalStateT . fmap (appEndo . foldMap Endo) . traverse f
+
+    -- Given a pair of indexed conversion postings, and a state consisting of lists of
+    -- costful and costless non-conversion postings, create a function which adds a conversion cost
+    -- to the posting which matches the conversion postings if necessary,
+    -- and tags the conversion and matched postings. Then update the state by removing the
+    -- matched postings. If there are no matching postings or too much ambiguity,
+    -- return an error string annotated with the conversion postings.
+    -- If the first argument is true, do a dry run instead: identify and tag
+    -- the costful and conversion postings, but don't add costs.
+    addCostsToPostings :: Bool -> (IdxPosting, IdxPosting)
+                        -> StateT ([IdxPosting], [IdxPosting]) (Either Text) (IdxPosting -> IdxPosting)
+    addCostsToPostings dryrun' ((n1, cp1), (n2, cp2)) = StateT $ \(costps, otherps) -> do
+        -- Get the two conversion posting amounts, if possible
+        ca1 <- conversionPostingAmountNoCost cp1
+        ca2 <- conversionPostingAmountNoCost cp2
+        let -- The function to add costs and tag postings in the indexed list of postings
+            transformPostingF np costp (n,p) =
+              (n, if | n == np            -> costp `postingAddTags` [("_price-matched","")]
+                     | n == n1 || n == n2 -> p     `postingAddTags` [("_conversion-matched","")]
+                     | otherwise          -> p)
+            -- All costful postings which match the conversion posting pair
+            matchingCostPs = mapMaybe (mapM $ costfulPostingIfMatchesBothAmounts ca1 ca2) costps
+            -- All other postings which match at least one of the conversion posting pair.
+            -- Add a corresponding cost to these postings, unless in dry run mode.
+            matchingOtherPs
+              | dryrun'   = [(n,(p, a)) | (n,p) <- otherps, let Just a = postingSingleAmount p]
+              | otherwise = mapMaybe (mapM $ addCostIfMatchesOneAmount ca1 ca2) otherps
+
+        -- Annotate any errors with the conversion posting pair
+        first (annotateWithPostings [cp1, cp2]) $
+            if -- If a single costful posting matches the conversion postings,
+               -- delete it from the list of costful postings in the state, delete the
+               -- first matching costless posting from the list of costless postings
+               -- in the state, and return the transformation function with the new state.
+               | [(np, (costp, _))] <- matchingCostPs
+               , Just newcostps <- deleteIdx np costps
+                   -> Right (transformPostingF np costp, (if dryrun' then costps else newcostps, otherps))
+               -- If no costful postings match the conversion postings, but some
+               -- of the costless postings match, check that the first such posting has a
+               -- different amount from all the others, and if so add a cost to it,
+               -- then delete it from the list of costless postings in the state,
+               -- and return the transformation function with the new state.
+               | [] <- matchingCostPs
+               , (np, (costp, amt)):nps <- matchingOtherPs
+               , not $ any (amountMatches amt . snd . snd) nps
+               , Just newotherps <- deleteIdx np otherps
+                   -> Right (transformPostingF np costp, (costps, if dryrun' then otherps else newotherps))
+               -- Otherwise it's too ambiguous to make a guess, so return an error.
+               | otherwise -> Left "There is not a unique posting which matches the conversion posting pair:"
+
+    -- If a posting with cost matches both the conversion amounts, return it along
+    -- with the matching amount which must be present in another non-conversion posting.
+    costfulPostingIfMatchesBothAmounts :: Amount -> Amount -> Posting -> Maybe (Posting, Amount)
+    costfulPostingIfMatchesBothAmounts a1 a2 p = do
+        a@Amount{aprice=Just _} <- postingSingleAmount p
+        if | amountMatches (-a1) a && amountMatches a2 (amountCost a) -> Just (p, -a2)
+           | amountMatches (-a2) a && amountMatches a1 (amountCost a) -> Just (p, -a1)
+           | otherwise -> Nothing
+
+    -- Add a cost to a posting if it matches (negative) one of the
+    -- supplied conversion amounts, adding the other amount as the cost.
+    addCostIfMatchesOneAmount :: Amount -> Amount -> Posting -> Maybe (Posting, Amount)
+    addCostIfMatchesOneAmount a1 a2 p = do
+        a <- postingSingleAmount p
+        let newp cost = p{pamount = mixedAmount a{aprice = Just $ TotalPrice cost}}
+        if | amountMatches (-a1) a -> Just (newp a2, a2)
+           | amountMatches (-a2) a -> Just (newp a1, a1)
+           | otherwise             -> Nothing
+
+    -- Get the single-commodity costless amount from a conversion posting, or raise an error.
+    conversionPostingAmountNoCost p = case postingSingleAmount p of
+        Just a@Amount{aprice=Nothing} -> Right a
+        Just Amount{aprice=Just _} -> Left $ annotateWithPostings [p] "Conversion postings must not have a cost:"
+        Nothing                    -> Left $ annotateWithPostings [p] "Conversion postings must have a single-commodity amount:"
+
+    amountMatches a b = acommodity a == acommodity b && aquantity a == aquantity b
+
+    -- Delete a posting from the indexed list of postings based on either its
+    -- index or its posting amount.
+    -- Note: traversing the whole list to delete a single match is generally not efficient,
+    -- but given that a transaction probably doesn't have more than four postings, it should
+    -- still be more efficient than using a Map or another data structure. Even monster
+    -- transactions with up to 10 postings, which are generally not a good
+    -- idea, are still too small for there to be an advantage.
+    -- XXX shouldn't assume transactions have few postings
+    deleteIdx n = deleteUniqueMatch ((n==) . fst)
+    deleteUniqueMatch p (x:xs) | p x       = if any p xs then Nothing else Just xs
+                               | otherwise = (x:) <$> deleteUniqueMatch p xs
+    deleteUniqueMatch _ []                 = Nothing
+    annotateWithPostings xs str = T.unlines $ str : postingsAsLines False xs
+
+-- Using the provided account types map, sort the given indexed postings
+-- into three lists of posting numbers (stored in two pairs), like so:
+-- (conversion postings, (costful postings, other postings)).
+-- A true first argument activates its secondary function: check that all
+-- conversion postings occur in adjacent pairs, otherwise return an error.
+partitionAndCheckConversionPostings :: Bool -> M.Map AccountName AccountType -> [IdxPosting] -> Either Text ( [(IdxPosting, IdxPosting)], ([IdxPosting], [IdxPosting]) )
+partitionAndCheckConversionPostings check acctTypes = fmap fst . foldrM select (([], ([], [])), Nothing)
+  where
+    select np@(_, p) ((cs, others@(ps, os)), Nothing)
+      | isConversion p = Right ((cs, others),      Just np)
+      | hasCost p      = Right ((cs, (np:ps, os)), Nothing)
+      | otherwise      = Right ((cs, (ps, np:os)), Nothing)
+    select np@(_, p) ((cs, others@(ps,os)), Just lst)
+      | isConversion p = Right (((lst, np):cs, others), Nothing)
+      | check          = Left "Conversion postings must occur in adjacent pairs"
+      | otherwise      = Right ((cs, (ps, np:os)), Nothing)
+    isConversion p = M.lookup (paccount p) acctTypes == Just Conversion
+    hasCost p = isJust $ aprice =<< postingSingleAmount p
+
+-- | Get a posting's amount if it is single-commodity.
+postingSingleAmount :: Posting -> Maybe Amount
+postingSingleAmount p = case amountsRaw (pamount p) of
+  [a] -> Just a
+  _   -> Nothing
 
 -- | Apply some account aliases to all posting account names in the transaction, as described by accountNameApplyAliases.
 -- This can fail due to a bad replacement pattern in a regular expression alias.
@@ -232,6 +383,13 @@ transactionMapPostingAmounts f  = transactionMapPostings (postingTransformAmount
 transactionFile :: Transaction -> FilePath
 transactionFile Transaction{tsourcepos} = sourceName $ fst tsourcepos
 
+-- Add transaction information to an error message.
+annotateErrorWithTransaction :: Transaction -> String -> String
+annotateErrorWithTransaction t s =
+  unlines [ sourcePosPairPretty $ tsourcepos t, s
+          , T.unpack . T.stripEnd $ showTransaction t
+          ]
+
 -- tests
 
 tests_Transaction :: TestTree
