@@ -1,4 +1,4 @@
-{-|
+  {-|
 
 Most data types are defined here to avoid import cycles.
 Here is an overview of the hledger data model:
@@ -49,7 +49,7 @@ import Data.Time.Clock.POSIX (POSIXTime)
 import Data.Time.LocalTime (LocalTime)
 import Data.Word (Word8)
 import Text.Blaze (ToMarkup(..))
-import Text.Megaparsec (SourcePos)
+import Text.Megaparsec (SourcePos(SourcePos), mkPos)
 
 import Hledger.Utils.Regex
 
@@ -81,8 +81,9 @@ type WeekDay = Int   -- 1-7
 -- (last|this|next) (day|week|month|quarter|year), where "this" means the period
 -- containing the reference date.
 data SmartDate
-  = SmartAssumeStart Year (Maybe (Month, Maybe MonthDay))
-  | SmartFromReference (Maybe Month) MonthDay
+  = SmartCompleteDate Day
+  | SmartAssumeStart Year (Maybe Month)         -- XXX improve these constructor names
+  | SmartFromReference (Maybe Month) MonthDay   --
   | SmartMonth Month
   | SmartRelative Integer SmartInterval
   deriving (Show)
@@ -91,7 +92,27 @@ data SmartInterval = Day | Week | Month | Quarter | Year deriving (Show)
 
 data WhichDate = PrimaryDate | SecondaryDate deriving (Eq,Show)
 
-data DateSpan = DateSpan (Maybe Day) (Maybe Day) deriving (Eq,Ord,Generic)
+-- | A date which is either exact or flexible.
+-- Flexible dates are allowed to be adjusted in certain situations.
+data EFDay = Exact Day | Flex Day deriving (Eq,Generic,Show)
+
+-- EFDay's Ord instance treats them like ordinary dates, ignoring exact/flexible.
+instance Ord EFDay where compare d1 d2 = compare (fromEFDay d1) (fromEFDay d2)
+
+-- instance Ord EFDay where compare = maCompare
+
+fromEFDay :: EFDay -> Day
+fromEFDay (Exact d) = d
+fromEFDay (Flex  d) = d
+
+modifyEFDay :: (Day -> Day) -> EFDay -> EFDay
+modifyEFDay f (Exact d) = Exact $ f d
+modifyEFDay f (Flex  d) = Flex  $ f d
+
+-- | A possibly open-ended span of time, from an optional inclusive start date
+-- to an optional exclusive end date. Each date can be either exact or flexible.
+-- An "exact date span" is a Datepan with exact start and end dates.
+data DateSpan = DateSpan (Maybe EFDay) (Maybe EFDay) deriving (Eq,Ord,Generic)
 
 instance Default DateSpan where def = DateSpan Nothing Nothing
 
@@ -162,6 +183,21 @@ instance Show AccountType where
   show Cash       = "C"
   show Conversion = "V"
 
+isBalanceSheetAccountType :: AccountType -> Bool
+isBalanceSheetAccountType t = t `elem` [
+  Asset,
+  Liability,
+  Equity,
+  Cash,
+  Conversion
+  ]
+
+isIncomeStatementAccountType :: AccountType -> Bool
+isIncomeStatementAccountType t = t `elem` [
+  Revenue,
+  Expense
+  ]
+
 -- | Check whether the first argument is a subtype of the second: either equal
 -- or one of the defined subtypes.
 isAccountSubtypeOf :: AccountType -> AccountType -> Bool
@@ -209,7 +245,7 @@ deriving instance Generic (DecimalRaw a)
 
 -- | An amount's per-unit or total cost/selling price in another
 -- commodity, as recorded in the journal entry eg with @ or @@.
--- Docs call this "transaction price". The amount is always positive.
+-- "Cost", formerly AKA "transaction price". The amount is always positive.
 data AmountPrice = UnitPrice !Amount | TotalPrice !Amount
   deriving (Eq,Ord,Generic,Show)
 
@@ -337,38 +373,14 @@ instance Show Status where -- custom show.. bad idea.. don't do it..
 -- at a certain point (posting date and parse order). They provide additional
 -- error checking and readability to a journal file.
 --
--- The 'BalanceAssertion' type is also used to represent balance assignments,
--- which instruct hledger what an account's balance should become at a certain
--- point.
+-- A balance assignments is an instruction to hledger to adjust an
+-- account's balance to a certain amount at a certain point.
 --
--- Different kinds of balance assertions are discussed eg on #290.
--- Variables include:
+-- The 'BalanceAssertion' type is used for representing both of these.
 --
--- - which postings are to be summed (real/virtual; unmarked/pending/cleared; this account/this account including subs)
---
--- - which commodities within the balance are to be checked
---
--- - whether to do a partial or a total check (disallowing other commodities)
---
--- I suspect we want:
---
--- 1. partial, subaccount-exclusive, Ledger-compatible assertions. Because
---    they're what we've always had, and removing them would break some
---    journals unnecessarily.  Implemented with = syntax.
---
--- 2. total assertions. Because otherwise assertions are a bit leaky.
---    Implemented with == syntax.
---
--- 3. subaccount-inclusive assertions. Because that's something folks need.
---    Not implemented.
---
--- 4. flexible assertions allowing custom criteria (perhaps arbitrary
---    queries). Because power users have diverse needs and want to try out
---    different schemes (assert cleared balances, assert balance from real or
---    virtual postings, etc.). Not implemented.
---
--- 5. multicommodity assertions, asserting the balance of multiple commodities
---    at once. Not implemented, requires #934.
+-- hledger supports multiple kinds of balance assertions/assignments,
+-- which differ in whether they refer to a single commodity or all commodities,
+-- and the (subaccount-)inclusive or exclusive account balance.
 --
 data BalanceAssertion = BalanceAssertion {
       baamount    :: Amount,    -- ^ the expected balance in a particular commodity
@@ -463,6 +475,7 @@ data PeriodicTransaction = PeriodicTransaction {
       ptinterval     :: Interval, -- ^ the interval at which this transaction recurs
       ptspan         :: DateSpan, -- ^ the (possibly unbounded) period during which this transaction recurs. Contains a whole number of intervals.
       --
+      ptsourcepos    :: (SourcePos, SourcePos),  -- ^ the file position where the period expression starts, and where the last posting ends
       ptstatus       :: Status,   -- ^ some of Transaction's fields
       ptcode         :: Text,
       ptdescription  :: Text,
@@ -475,6 +488,7 @@ nullperiodictransaction = PeriodicTransaction{
       ptperiodexpr   = ""
      ,ptinterval     = def
      ,ptspan         = def
+     ,ptsourcepos    = (SourcePos "" (mkPos 1) (mkPos 1), SourcePos "" (mkPos 1) (mkPos 1))
      ,ptstatus       = Unmarked
      ,ptcode         = ""
      ,ptdescription  = ""
@@ -490,7 +504,9 @@ data TimeclockEntry = TimeclockEntry {
       tlcode        :: TimeclockCode,
       tldatetime    :: LocalTime,
       tlaccount     :: AccountName,
-      tldescription :: Text
+      tldescription :: Text,
+      tlcomment     :: Text,
+      tltags        :: [Tag]
     } deriving (Eq,Ord,Generic)
 
 -- | A market price declaration made by the journal format's P directive.
@@ -535,6 +551,7 @@ data Journal = Journal {
   ,jincludefilestack      :: [FilePath]
   -- principal data
   ,jdeclaredpayees        :: [(Payee,PayeeDeclarationInfo)]         -- ^ Payees declared by payee directives, in parse order (after journal finalisation)
+  ,jdeclaredtags          :: [(TagName,TagDeclarationInfo)]         -- ^ Tags declared by tag directives, in parse order (after journal finalisation)
   ,jdeclaredaccounts      :: [(AccountName,AccountDeclarationInfo)] -- ^ Accounts declared by account directives, in parse order (after journal finalisation)
   ,jdeclaredaccounttags   :: M.Map AccountName [Tag]                -- ^ Accounts which have tags declared in their directives, and those tags. (Does not include parents' tags.)
   ,jdeclaredaccounttypes  :: M.Map AccountType [AccountName]        -- ^ Accounts whose type has been explicitly declared in their account directives, grouped by type.
@@ -577,6 +594,15 @@ nullpayeedeclarationinfo = PayeeDeclarationInfo {
   ,pditags             = []
 }
 
+-- | Extra information found in a tag directive.
+newtype TagDeclarationInfo = TagDeclarationInfo {
+   tdicomment :: Text   -- ^ any comment lines following the tag directive. No tags allowed here.
+} deriving (Eq,Show,Generic)
+
+nulltagdeclarationinfo = TagDeclarationInfo {
+   tdicomment          = ""
+}
+
 -- | Extra information about an account that can be derived from
 -- its account directive (and the other account directives).
 data AccountDeclarationInfo = AccountDeclarationInfo {
@@ -584,12 +610,14 @@ data AccountDeclarationInfo = AccountDeclarationInfo {
   ,aditags             :: [Tag]  -- ^ tags extracted from the account comment, if any
   ,adideclarationorder :: Int    -- ^ the order in which this account was declared,
                                  --   relative to other account declarations, during parsing (1..)
+  ,adisourcepos        :: SourcePos  -- ^ source file and position
 } deriving (Eq,Show,Generic)
 
 nullaccountdeclarationinfo = AccountDeclarationInfo {
    adicomment          = ""
   ,aditags             = []
   ,adideclarationorder = 0
+  ,adisourcepos        = SourcePos "" (mkPos 1) (mkPos 1)
 }
 
 -- | An account, with its balances, parent/subaccount relationships, etc.
