@@ -24,6 +24,7 @@ module Hledger.Data.Valuation (
   ,mixedAmountGainAtDate
   ,marketPriceReverse
   ,priceDirectiveToMarketPrice
+  ,amountPriceDirectiveFromCost
   -- ,priceLookup
   ,tests_Valuation
 )
@@ -96,6 +97,22 @@ priceDirectiveToMarketPrice PriceDirective{..} =
              , mprate = aquantity pdamount
              }
 
+-- | Make one or more `MarketPrice` from an 'Amount' and its price directives.
+amountPriceDirectiveFromCost :: Day -> Amount -> Maybe PriceDirective
+amountPriceDirectiveFromCost d amt@Amount{acommodity=fromcomm, aquantity=fromq} = case aprice amt of
+    Just (UnitPrice pa)               -> Just $ pd{pdamount=pa}
+    Just (TotalPrice pa) | fromq /= 0 -> Just $ pd{pdamount=fromq `divideAmountExtraPrecision` pa}
+    _                                 -> Nothing
+  where
+    pd = PriceDirective{pddate = d, pdcommodity = fromcomm, pdamount = nullamt}
+
+    divideAmountExtraPrecision n a = (n `divideAmount` a) { astyle = style' }
+      where
+        style' = (astyle a) { asprecision = precision' }
+        precision' = case asprecision (astyle a) of
+                          NaturalPrecision -> NaturalPrecision
+                          Precision p      -> Precision $ (numDigitsInt $ truncate n) + p
+
 ------------------------------------------------------------------------------
 -- Converting things to value
 
@@ -176,7 +193,7 @@ amountValueAtDate priceoracle styles mto d a =
                                       -- Make default display style use precision 2 instead of 0 ?
                                       -- Leave as is for now; mentioned in manual.
       styleAmount styles
-      amount{acommodity=comm, aquantity=rate * aquantity a}
+      nullamt{acommodity=comm, aquantity=rate * aquantity a}
 
 -- | Calculate the gain of each component amount, that is the difference
 -- between the valued amount and the value of the cost basis (see
@@ -227,7 +244,7 @@ priceLookup makepricegraph d from mto =
               ,pgEdgesRev=allprices
               ,pgDefaultValuationCommodities=defaultdests
               } =
-      traceAt 1 ("valuation date: "++show d) $ makepricegraph d
+      traceOrLogAt 1 ("valuation date: "++show d) $ makepricegraph d
     mto' = mto <|> mdefaultto
       where
         mdefaultto = dbg1 ("default valuation commodity for "++T.unpack from) $
@@ -241,10 +258,10 @@ priceLookup makepricegraph d from mto =
         -- according to the rules described in makePriceGraph.
         let msg = printf "seeking %s to %s price" (showCommoditySymbol from) (showCommoditySymbol to)
         in case 
-          (traceAt 2 (msg++" using forward prices") $ 
+          (traceOrLogAt 2 (msg++" using forward prices") $ 
             pricesShortestPath from to forwardprices)
           <|> 
-          (traceAt 2 (msg++" using forward and reverse prices") $ 
+          (traceOrLogAt 2 (msg++" using forward and reverse prices") $ 
             pricesShortestPath from to allprices)
         of
           Nothing -> Nothing
@@ -318,7 +335,7 @@ pricesShortestPath start end edges =
       case concatMap extend paths of
         [] -> Nothing 
         _ | pathlength > maxpathlength -> 
-          trace ("gave up searching for a price chain at length "++show maxpathlength++", please report a bug")
+          traceOrLog ("gave up searching for a price chain at length "++show maxpathlength++", please report a bug")
           Nothing
           where 
             pathlength = 2 + maybe 0 (length . fst) (headMay paths)
