@@ -4,7 +4,7 @@
 {-|
 
 A reader for hledger's journal file format
-(<http://hledger.org/MANUAL.html#the-journal-file>).  hledger's journal
+(<http://hledger.org/hledger.html#the-journal-file>).  hledger's journal
 format is a compatible subset of c++ ledger's
 (<http://ledger-cli.org/3.0/doc/ledger3.html#Journal-Format>), so this
 reader should handle many ledger files as well. Example:
@@ -79,7 +79,7 @@ import Control.Monad.Except (ExceptT(..), runExceptT)
 import Control.Monad.State.Strict (evalStateT,get,modify',put)
 import Control.Monad.Trans.Class (lift)
 import Data.Char (toLower)
-import Data.Either (isRight)
+import Data.Either (isRight, lefts)
 import qualified Data.Map.Strict as M
 import Data.Text (Text)
 import Data.String
@@ -100,9 +100,10 @@ import Hledger.Data
 import Hledger.Read.Common
 import Hledger.Utils
 
-import qualified Hledger.Read.TimedotReader as TimedotReader (reader)
-import qualified Hledger.Read.TimeclockReader as TimeclockReader (reader)
 import qualified Hledger.Read.CsvReader as CsvReader (reader)
+import qualified Hledger.Read.RulesReader as RulesReader (reader)
+import qualified Hledger.Read.TimeclockReader as TimeclockReader (reader)
+import qualified Hledger.Read.TimedotReader as TimedotReader (reader)
 
 --- ** doctest setup
 -- $setup
@@ -113,7 +114,7 @@ import qualified Hledger.Read.CsvReader as CsvReader (reader)
 -- | Run a journal parser in some monad. See also: parseWithState.
 runJournalParser, rjp
   :: Monad m
-  => JournalParser m a -> Text -> m (Either (ParseErrorBundle Text CustomErr) a)
+  => JournalParser m a -> Text -> m (Either HledgerParseErrors a)
 runJournalParser p = runParserT (evalStateT p nulljournal) ""
 rjp = runJournalParser
 
@@ -122,7 +123,7 @@ runErroringJournalParser, rejp
   :: Monad m
   => ErroringJournalParser m a
   -> Text
-  -> m (Either FinalParseError (Either (ParseErrorBundle Text CustomErr) a))
+  -> m (Either FinalParseError (Either HledgerParseErrors a))
 runErroringJournalParser p t =
   runExceptT $ runParserT (evalStateT p nulljournal) "" t
 rejp = runErroringJournalParser
@@ -137,6 +138,7 @@ readers' = [
   reader
  ,TimeclockReader.reader
  ,TimedotReader.reader
+ ,RulesReader.reader
  ,CsvReader.reader
 --  ,LedgerReader.reader
  ]
@@ -168,7 +170,7 @@ type PrefixedFilePath = FilePath
 -- split that off. Eg "csv:-" -> (Just "csv", "-").
 splitReaderPrefix :: PrefixedFilePath -> (Maybe String, FilePath)
 splitReaderPrefix f =
-  headDef (Nothing, f)
+  headDef (Nothing, f) $
   [(Just r, drop (length r + 1) f) | r <- readerNames, (r++":") `isPrefixOf` f]
 
 --- ** reader
@@ -185,7 +187,7 @@ reader = Reader
 -- | Parse and post-process a "Journal" from hledger's journal file
 -- format, or give an error.
 parse :: InputOpts -> FilePath -> Text -> ExceptT String IO Journal
-parse iopts = parseAndFinaliseJournal journalp' iopts
+parse iopts f = parseAndFinaliseJournal journalp' iopts f
   where
     journalp' = do
       -- reverse parsed aliases to ensure that they are applied in order given on commandline
@@ -199,7 +201,7 @@ parse iopts = parseAndFinaliseJournal journalp' iopts
 -- which should be finalised/validated before use.
 --
 -- >>> rejp (journalp <* eof) "2015/1/1\n a  0\n"
--- Right (Right Journal  with 1 transactions, 1 accounts)
+-- Right (Right Journal (unknown) with 1 transactions, 1 accounts)
 --
 journalp :: MonadIO m => ErroringJournalParser m ParsedJournal
 journalp = do
@@ -226,27 +228,42 @@ addJournalItemP =
 --- *** directives
 
 -- | Parse any journal directive and update the parse state accordingly.
--- Cf http://hledger.org/manual.html#directives,
+-- Cf http://hledger.org/hledger.html#directives,
 -- http://ledger-cli.org/3.0/doc/ledger3.html#Command-Directives
 directivep :: MonadIO m => ErroringJournalParser m ()
 directivep = (do
-  optional $ char '!'
+  optional $ oneOf ['!','@']
   choice [
     includedirectivep
    ,aliasdirectivep
    ,endaliasesdirectivep
    ,accountdirectivep
    ,applyaccountdirectivep
+   ,applyfixeddirectivep
+   ,applytagdirectivep
+   ,assertdirectivep
+   ,bucketdirectivep
+   ,capturedirectivep
+   ,checkdirectivep
+   ,commandlineflagdirectivep
    ,commoditydirectivep
-   ,endapplyaccountdirectivep
-   ,payeedirectivep
-   ,tagdirectivep
-   ,endtagdirectivep
+   ,commodityconversiondirectivep
+   ,decimalmarkdirectivep
    ,defaultyeardirectivep
    ,defaultcommoditydirectivep
-   ,commodityconversiondirectivep
+   ,definedirectivep
+   ,endapplyaccountdirectivep
+   ,endapplyfixeddirectivep
+   ,endapplytagdirectivep
+   ,endapplyyeardirectivep
+   ,endtagdirectivep
+   ,evaldirectivep
+   ,exprdirectivep
    ,ignoredpricecommoditydirectivep
-   ,decimalmarkdirectivep
+   ,payeedirectivep
+   ,pythondirectivep
+   ,tagdirectivep
+   ,valuedirectivep
    ]
   ) <?> "directive"
 
@@ -261,8 +278,8 @@ includedirectivep = do
   prefixedglob <- T.unpack <$> takeWhileP Nothing (/= '\n') -- don't consume newline yet
   parentoff <- getOffset
   parentpos <- getSourcePos
-  let (mprefix,glob) = splitReaderPrefix prefixedglob
-  paths <- getFilePaths parentoff parentpos glob
+  let (mprefix,glb) = splitReaderPrefix prefixedglob
+  paths <- getFilePaths parentoff parentpos glb
   let prefixedpaths = case mprefix of
         Nothing  -> paths
         Just fmt -> map ((fmt++":")++) paths
@@ -298,20 +315,40 @@ includedirectivep = do
       when (filepath `elem` parentfilestack) $
         Fail.fail ("Cyclic include: " ++ filepath)
 
-      childInput <- lift $ readFilePortably filepath
-                            `orRethrowIOError` (show parentpos ++ " reading " ++ filepath)
+      childInput <-
+        traceOrLogAt 6 ("parseChild: "++takeFileName filepath) $
+        lift $ readFilePortably filepath
+          `orRethrowIOError` (show parentpos ++ " reading " ++ filepath)
       let initChildj = newJournalWithParseStateFrom filepath parentj
 
-      -- Choose a reader/format based on the file path, or fall back
-      -- on journal. Duplicating readJournal a bit here.
+      -- Choose a reader/parser based on the file path prefix or file extension,
+      -- defaulting to JournalReader. Duplicating readJournal a bit here.
       let r = fromMaybe reader $ findReader Nothing (Just prefixedpath)
           parser = rParser r
-      dbg6IO "trying reader" (rFormat r)
+      dbg6IO "parseChild: trying reader" (rFormat r)
+
+      -- Parse the file (of whichever format) to a Journal, with file path and source text attached.
       updatedChildj <- journalAddFile (filepath, childInput) <$>
                         parseIncludeFile parser initChildj filepath childInput
 
-      -- discard child's parse info,  combine other fields
-      put $ updatedChildj <> parentj
+      -- Merge this child journal into the parent journal
+      -- (with debug logging for troubleshooting account display order).
+      -- The parent journal is the second argument to journalConcat; this means
+      -- its parse state is kept, and its lists are appended to child's (which
+      -- ultimately produces the right list order, because parent's and child's
+      -- lists are in reverse order at this stage. Cf #1909).
+      let
+        parentj' =
+          dbgJournalAcctDeclOrder ("parseChild: child " <> childfilename <> " acct decls: ") updatedChildj
+          `journalConcat`
+          dbgJournalAcctDeclOrder ("parseChild: parent " <> parentfilename <> " acct decls: ") parentj
+
+          where
+            childfilename = takeFileName filepath
+            parentfilename = maybe "(unknown)" takeFileName $ headMay $ jincludefilestack parentj  -- XXX more accurate than journalFilePath for some reason
+
+      -- Update the parse state.
+      put parentj'
 
     newJournalWithParseStateFrom :: FilePath -> Journal -> Journal
     newJournalWithParseStateFrom filepath j = nulljournal{
@@ -340,12 +377,14 @@ orRethrowIOError io msg = do
 accountdirectivep :: JournalParser m ()
 accountdirectivep = do
   off <- getOffset -- XXX figure out a more precise position later
+  pos <- getSourcePos
 
   string "account"
   lift skipNonNewlineSpaces1
 
   -- the account name, possibly modified by preceding alias or apply account directives
-  acct <- modifiedaccountnamep
+  acct <- (notFollowedBy (char '(' <|> char '[') <?> "account name without brackets") >>
+          modifiedaccountnamep
 
   -- maybe a comment, on this and/or following lines
   (cmt, tags) <- lift transactioncommentp
@@ -359,7 +398,7 @@ accountdirectivep = do
     metype = parseAccountTypeCode <$> lookup accountTypeTagName tags
 
   -- update the journal
-  addAccountDeclaration (acct, cmt, tags)
+  addAccountDeclaration (acct, cmt, tags, pos)
   unless (null tags) $ addDeclaredAccountTags acct tags
   case metype of
     Nothing         -> return ()
@@ -392,15 +431,16 @@ parseAccountTypeCode s =
             T.intercalate ", " ["A","L","E","R","X","C","V","Asset","Liability","Equity","Revenue","Expense","Cash","Conversion"]
 
 -- Add an account declaration to the journal, auto-numbering it.
-addAccountDeclaration :: (AccountName,Text,[Tag]) -> JournalParser m ()
-addAccountDeclaration (a,cmt,tags) =
+addAccountDeclaration :: (AccountName,Text,[Tag],SourcePos) -> JournalParser m ()
+addAccountDeclaration (a,cmt,tags,pos) = do
   modify' (\j ->
              let
                decls = jdeclaredaccounts j
                d     = (a, nullaccountdeclarationinfo{
                               adicomment          = cmt
                              ,aditags             = tags
-                             ,adideclarationorder = length decls + 1
+                             ,adideclarationorder = length decls + 1  -- gets renumbered when Journals are finalised or merged
+                             ,adisourcepos        = pos
                              })
              in
                j{jdeclaredaccounts = d:decls})
@@ -416,6 +456,13 @@ addPayeeDeclaration (p, cmt, tags) =
                     ,pditags    = tags
                     })
 
+-- Add a tag declaration to the journal.
+addTagDeclaration :: (TagName,Text) -> JournalParser m ()
+addTagDeclaration (t, cmt) =
+  modify' (\j@Journal{jdeclaredtags} -> j{jdeclaredtags=tagandinfo:jdeclaredtags})
+  where
+    tagandinfo = (t, nulltagdeclarationinfo{tdicomment=cmt})
+
 indentedlinep :: JournalParser m String
 indentedlinep = lift skipNonNewlineSpaces1 >> (rstrip <$> lift restofline)
 
@@ -438,8 +485,8 @@ commoditydirectiveonelinep = do
     string "commodity"
     lift skipNonNewlineSpaces1
     off <- getOffset
-    amount <- amountp
-    pure $ (off, amount)
+    amt <- amountp
+    pure $ (off, amt)
   lift skipNonNewlineSpaces
   _ <- lift followingcommentp
   let comm = Commodity{csymbol=acommodity, cformat=Just $ dbg6 "style from commodity directive" astyle}
@@ -467,8 +514,10 @@ commoditydirectivemultilinep = do
   lift skipNonNewlineSpaces1
   sym <- lift commoditysymbolp
   _ <- lift followingcommentp
-  mformat <- lastMay <$> many (indented $ formatdirectivep sym)
-  let comm = Commodity{csymbol=sym, cformat=mformat}
+  -- read all subdirectives, saving format subdirectives as Lefts
+  subdirectives <- many $ indented (eitherP (formatdirectivep sym) (lift restofline))
+  let mfmt = lastMay $ lefts subdirectives
+  let comm = Commodity{csymbol=sym, cformat=mfmt}
   modify' (\j -> j{jcommodities=M.insert sym comm $ jcommodities j})
   where
     indented = (lift skipNonNewlineSpaces1 >>)
@@ -490,11 +539,40 @@ formatdirectivep expectedsym = do
     else customFailure $ parseErrorAt off $
          printf "commodity directive symbol \"%s\" and format directive symbol \"%s\" should be the same" expectedsym acommodity
 
+-- More Ledger directives, ignore for now:
+-- apply fixed, apply tag, assert, bucket, A, capture, check, define, expr
+applyfixeddirectivep, endapplyfixeddirectivep, applytagdirectivep, endapplytagdirectivep,
+  assertdirectivep, bucketdirectivep, capturedirectivep, checkdirectivep, 
+  endapplyyeardirectivep, definedirectivep, exprdirectivep, valuedirectivep,
+  evaldirectivep, pythondirectivep, commandlineflagdirectivep
+  :: JournalParser m ()
+applyfixeddirectivep    = do string "apply fixed" >> lift restofline >> return ()
+endapplyfixeddirectivep = do string "end apply fixed" >> lift restofline >> return ()
+applytagdirectivep      = do string "apply tag" >> lift restofline >> return ()
+endapplytagdirectivep   = do string "end apply tag" >> lift restofline >> return ()
+endapplyyeardirectivep  = do string "end apply year" >> lift restofline >> return ()
+assertdirectivep        = do string "assert"  >> lift restofline >> return ()
+bucketdirectivep        = do string "A " <|> string "bucket " >> lift restofline >> return ()
+capturedirectivep       = do string "capture" >> lift restofline >> return ()
+checkdirectivep         = do string "check"   >> lift restofline >> return ()
+definedirectivep        = do string "define"  >> lift restofline >> return ()
+exprdirectivep          = do string "expr"    >> lift restofline >> return ()
+valuedirectivep         = do string "value"   >> lift restofline >> return ()
+evaldirectivep          = do string "eval"   >> lift restofline >> return ()
+commandlineflagdirectivep = do string "--" >> lift restofline >> return ()
+pythondirectivep = do
+  string "python" >> lift restofline
+  many $ indentedline <|> blankline
+  return ()
+  where
+    indentedline = lift skipNonNewlineSpaces1 >> lift restofline
+    blankline = lift skipNonNewlineSpaces >> newline >> return "" <?> "blank line"
+
 keywordp :: String -> JournalParser m ()
-keywordp = (() <$) . string . fromString
+keywordp = void . string . fromString
 
 spacesp :: JournalParser m ()
-spacesp = () <$ lift skipNonNewlineSpaces1
+spacesp = void $ lift skipNonNewlineSpaces1
 
 -- | Backtracking parser similar to string, but allows varying amount of space between words
 keywordsp :: String -> JournalParser m ()
@@ -529,15 +607,23 @@ tagdirectivep :: JournalParser m ()
 tagdirectivep = do
   string "tag" <?> "tag directive"
   lift skipNonNewlineSpaces1
-  _ <- lift $ some nonspace
-  lift restofline
+  tagname <- lift $ T.pack <$> some nonspace
+  (comment, _) <- lift transactioncommentp
+  skipMany indentedlinep
+  addTagDeclaration (tagname,comment)
   return ()
 
+-- end tag or end apply tag
 endtagdirectivep :: JournalParser m ()
-endtagdirectivep = do
-  (keywordsp "end tag" <|> keywordp "pop") <?> "end tag or pop directive"
-  lift restofline
+endtagdirectivep = (do
+  string "end"
+  lift skipNonNewlineSpaces1
+  optional $ string "apply" >> lift skipNonNewlineSpaces1
+  string "tag"
+  lift skipNonNewlineSpaces
+  eol
   return ()
+  ) <?> "end tag or end apply tag directive"
 
 payeedirectivep :: JournalParser m ()
 payeedirectivep = do
@@ -545,12 +631,13 @@ payeedirectivep = do
   lift skipNonNewlineSpaces1
   payee <- lift $ T.strip <$> noncommenttext1p
   (comment, tags) <- lift transactioncommentp
+  skipMany indentedlinep
   addPayeeDeclaration (payee, comment, tags)
   return ()
 
 defaultyeardirectivep :: JournalParser m ()
 defaultyeardirectivep = do
-  char 'Y' <?> "default year"
+  (string "Y" <|> string "year" <|> string "apply year") <?> "default year"
   lift skipNonNewlineSpaces
   setYear =<< lift yearp
 
@@ -632,12 +719,11 @@ transactionmodifierp = do
 -- relative to Y/1/1. If not, they are calculated related to today as usual.
 periodictransactionp :: MonadIO m => JournalParser m PeriodicTransaction
 periodictransactionp = do
+  startpos <- getSourcePos
 
   -- first line
   char '~' <?> "periodic transaction"
   lift $ skipNonNewlineSpaces
-  -- a period expression
-  off <- getOffset
 
   -- if there's a default year in effect, use Y/1/1 as base for partial/relative dates
   today <- liftIO getCurrentDay
@@ -652,7 +738,7 @@ periodictransactionp = do
   -- first parsing with 'singlespacedtextp', then "re-parsing" with
   -- 'periodexprp' saves 'periodexprp' from having to respect the single-
   -- and double-space parsing rules
-  (interval, span) <- lift $ reparseExcerpt periodExcerpt $ do
+  (interval, spn) <- lift $ reparseExcerpt periodExcerpt $ do
     pexp <- periodexprp refdate
     (<|>) eof $ do
       offset1 <- getOffset
@@ -664,11 +750,6 @@ periodictransactionp = do
         <> "\na double space is required between period expression and description/comment"
     pure pexp
 
-  -- In periodic transactions, the period expression has an additional constraint:
-  case checkPeriodicTransactionStartDate interval span periodtxt of
-    Just e -> customFailure $ parseErrorAt off e
-    Nothing -> pure ()
-
   status <- lift statusp <?> "cleared status"
   code <- lift codep <?> "transaction code"
   description <- lift $ T.strip <$> descriptionp
@@ -676,10 +757,14 @@ periodictransactionp = do
   -- next lines; use same year determined above
   postings <- postingsp (Just $ first3 $ toGregorian refdate)
 
+  endpos <- getSourcePos
+  let sourcepos = (startpos, endpos)
+
   return $ nullperiodictransaction{
      ptperiodexpr=periodtxt
     ,ptinterval=interval
-    ,ptspan=span
+    ,ptspan=spn
+    ,ptsourcepos=sourcepos
     ,ptstatus=status
     ,ptcode=code
     ,ptdescription=description
@@ -745,7 +830,7 @@ postingphelper isPostingRule mTransactionYear = do
     let (ptype, account') = (accountNamePostingType account, textUnbracket account)
     lift skipNonNewlineSpaces
     mult <- if isPostingRule then multiplierp else pure False
-    amount <- optional $ amountpwithmultiplier mult
+    amt <- optional $ amountp' mult
     lift skipNonNewlineSpaces
     massertion <- optional balanceassertionp
     lift skipNonNewlineSpaces
@@ -755,7 +840,7 @@ postingphelper isPostingRule mTransactionYear = do
             , pdate2=mdate2
             , pstatus=status
             , paccount=account'
-            , pamount=maybe missingmixedamt mixedAmount amount
+            , pamount=maybe missingmixedamt mixedAmount amt
             , pcomment=comment
             , ptype=ptype
             , ptags=tags
@@ -794,8 +879,8 @@ tests_JournalReader = testGroup "JournalReader" [
     ]
   ,testCase "datetimep" $ do
      let
-       good = assertParse datetimep
-       bad  = (\t -> assertParseError datetimep t "")
+       good  = assertParse datetimep
+       bad t = assertParseError datetimep t ""
      good "2011/1/1 00:00"
      good "2011/1/1 23:59:59"
      bad "2011/1/1"
@@ -815,7 +900,8 @@ tests_JournalReader = testGroup "JournalReader" [
       nullperiodictransaction {
          ptperiodexpr  = "monthly from 2018/6"
         ,ptinterval    = Months 1
-        ,ptspan        = DateSpan (Just $ fromGregorian 2018 6 1) Nothing
+        ,ptspan        = DateSpan (Just $ Flex $ fromGregorian 2018 6 1) Nothing
+        ,ptsourcepos   = (SourcePos "" (mkPos 1) (mkPos 1), SourcePos "" (mkPos 2) (mkPos 1))
         ,ptdescription = ""
         ,ptcomment     = "In 2019 we will change this\n"
         }
@@ -825,7 +911,8 @@ tests_JournalReader = testGroup "JournalReader" [
       nullperiodictransaction {
          ptperiodexpr  = "monthly from 2018/6"
         ,ptinterval    = Months 1
-        ,ptspan        = DateSpan (Just $ fromGregorian 2018 6 1) Nothing
+        ,ptspan        = DateSpan (Just $ Flex $ fromGregorian 2018 6 1) Nothing
+        ,ptsourcepos   = (SourcePos "" (mkPos 1) (mkPos 1), SourcePos "" (mkPos 2) (mkPos 1))
         ,ptdescription = "In 2019 we will change this"
         ,ptcomment     = ""
         }
@@ -836,6 +923,7 @@ tests_JournalReader = testGroup "JournalReader" [
          ptperiodexpr  = "monthly"
         ,ptinterval    = Months 1
         ,ptspan        = DateSpan Nothing Nothing
+        ,ptsourcepos   = (SourcePos "" (mkPos 1) (mkPos 1), SourcePos "" (mkPos 2) (mkPos 1))
         ,ptdescription = "Next year blah blah"
         ,ptcomment     = ""
         }
@@ -845,7 +933,8 @@ tests_JournalReader = testGroup "JournalReader" [
       nullperiodictransaction {
          ptperiodexpr  = "2019-01-04"
         ,ptinterval    = NoInterval
-        ,ptspan        = DateSpan (Just $ fromGregorian 2019 1 4) (Just $ fromGregorian 2019 1 5)
+        ,ptspan        = DateSpan (Just $ Exact $ fromGregorian 2019 1 4) (Just $ Exact $ fromGregorian 2019 1 5)
+        ,ptsourcepos   = (SourcePos "" (mkPos 1) (mkPos 1), SourcePos "" (mkPos 2) (mkPos 1))
         ,ptdescription = ""
         ,ptcomment     = ""
         }
@@ -1010,6 +1099,7 @@ tests_JournalReader = testGroup "JournalReader" [
         [("a:b", AccountDeclarationInfo{adicomment          = "type:asset\n"
                                        ,aditags             = [("type","asset")]
                                        ,adideclarationorder = 1
+                                       ,adisourcepos        = fst nullsourcepos
                                        })
         ]
       ]
@@ -1053,7 +1143,7 @@ tests_JournalReader = testGroup "JournalReader" [
 
   ,testCase "endtagdirectivep" $ do
       assertParse endtagdirectivep "end tag \n"
-      assertParse endtagdirectivep "pop \n"
+      assertParse endtagdirectivep "end apply tag \n"
 
   ,testGroup "journalp" [
     testCase "empty file" $ assertParseEqE journalp "" nulljournal
