@@ -28,7 +28,6 @@ inc.client1   .... .... ..
 
 --- ** language
 {-# LANGUAGE OverloadedStrings #-}
-{-# LANGUAGE PackageImports #-}
 
 --- ** exports
 module Hledger.Read.TimedotReader (
@@ -41,10 +40,9 @@ where
 
 --- ** imports
 import Control.Monad
-import Control.Monad.Except (ExceptT)
+import Control.Monad.Except (ExceptT, liftEither)
 import Control.Monad.State.Strict
 import Data.Char (isSpace)
-import Data.List (foldl')
 import Data.Text (Text)
 import qualified Data.Text as T
 import Data.Time (Day)
@@ -71,7 +69,9 @@ reader = Reader
 
 -- | Parse and post-process a "Journal" from the timedot format, or give an error.
 parse :: InputOpts -> FilePath -> Text -> ExceptT String IO Journal
-parse = parseAndFinaliseJournal' timedotp
+parse iopts fp t = initialiseAndParseJournal timedotp iopts fp t
+                   >>= liftEither . journalApplyAliases (aliasesFromOpts iopts)
+                   >>= journalFinalise iopts fp t
 
 --- ** utilities
 
@@ -112,7 +112,7 @@ preamblep = do
   many $ notFollowedBy datelinep >> (lift $ emptyorcommentlinep "#;*")
   lift $ traceparse' "preamblep"
 
--- | Parse timedot day entries to zero or more time transactions for that day.
+-- | Parse timedot day entries to multi-posting time transactions for that day.
 -- @
 -- 2020/2/1 optional day description
 -- fos.haskell  .... ..
@@ -122,23 +122,32 @@ preamblep = do
 dayp :: JournalParser m ()
 dayp = label "timedot day entry" $ do
   lift $ traceparse "dayp"
-  (d,desc) <- datelinep
+  pos <- getSourcePos
+  (date,desc,comment,tags) <- datelinep
   commentlinesp
-  ts <- many $ entryp <* commentlinesp
-  modify' $ addTransactions $ map (\t -> t{tdate=d, tdescription=desc}) ts
-  lift $ traceparse' "dayp"
-  where
-    addTransactions :: [Transaction] -> Journal -> Journal
-    addTransactions ts j = foldl' (flip ($)) j (map addTransaction ts)
+  ps <- many $ timedotentryp <* commentlinesp
+  endpos <- getSourcePos
+  -- lift $ traceparse' "dayp end"
+  let t = txnTieKnot $ nulltransaction{
+    tsourcepos   = (pos, endpos),
+    tdate        = date,
+    tstatus      = Cleared,
+    tdescription = desc,
+    tcomment     = comment,
+    ttags        = tags,
+    tpostings    = ps
+    }
+  modify' $ addTransaction t
 
-datelinep :: JournalParser m (Day,Text)
+datelinep :: JournalParser m (Day,Text,Text,[Tag])
 datelinep = do
   lift $ traceparse "datelinep"
   lift $ optional orgheadingprefixp
-  d <- datep
-  desc <- strip <$> lift restofline
-  lift $ traceparse' "datelinep"
-  return (d, T.pack desc)
+  date <- datep
+  desc <- T.strip <$> lift descriptionp
+  (comment, tags) <- lift transactioncommentp
+  -- lift $ traceparse' "datelinep end"
+  return (date, desc, comment, tags)
 
 -- | Zero or more empty lines or hash/semicolon comment lines
 -- or org headlines which do not start a new day.
@@ -163,40 +172,41 @@ orgheadingprefixp = do
 -- @
 -- fos.haskell  .... ..
 -- @
-entryp :: JournalParser m Transaction
-entryp = do
-  lift $ traceparse "entryp"
-  pos <- getSourcePos
+timedotentryp :: JournalParser m Posting
+timedotentryp = do
+  lift $ traceparse "timedotentryp"
   notFollowedBy datelinep
   lift $ optional $ choice [orgheadingprefixp, skipNonNewlineSpaces1]
   a <- modifiedaccountnamep
   lift skipNonNewlineSpaces
-  hrs <-
-    try (lift followingcommentp >> return 0)
-    <|> (durationp <*
-         (try (lift followingcommentp) <|> (newline >> return "")))
+  (hours, comment, tags) <-
+    try (do
+      (c,ts) <- lift transactioncommentp  -- or postingp, but let's not bother supporting date:/date2:
+      return (0, c, ts)
+    )
+    <|> (do
+      h <- lift durationp
+      (c,ts) <- try (lift transactioncommentp) <|> (newline >> return ("",[]))
+      return (h,c,ts)
+    )
   mcs <- getDefaultCommodityAndStyle
   let 
     (c,s) = case mcs of
       Just (defc,defs) -> (defc, defs{asprecision=max (asprecision defs) (Precision 2)})
       _ -> ("", amountstyle{asprecision=Precision 2})
-    t = nulltransaction{
-          tsourcepos = (pos, pos),
-          tstatus    = Cleared,
-          tpostings  = [
-            nullposting{paccount=a
-                      ,pamount=mixedAmount $ nullamt{acommodity=c, aquantity=hrs, astyle=s}
+  -- lift $ traceparse' "timedotentryp end"
+  return $ nullposting{paccount=a
+                      ,pamount=mixedAmount $ nullamt{acommodity=c, aquantity=hours, astyle=s}
                       ,ptype=VirtualPosting
-                      ,ptransaction=Just t
+                      ,pcomment=comment
+                      ,ptags=tags
                       }
-            ]
-          }
-  lift $ traceparse' "entryp"
-  return t
 
-durationp :: JournalParser m Quantity
+type Hours = Quantity
+
+durationp :: TextParser m Hours
 durationp = do
-  lift $ traceparse "durationp"
+  traceparse "durationp"
   try numericquantityp <|> dotquantityp
     -- <* traceparse' "durationp"
 
@@ -209,12 +219,12 @@ durationp = do
 -- 1.5h
 -- 90m
 -- @
-numericquantityp :: JournalParser m Quantity
+numericquantityp :: TextParser m Hours
 numericquantityp = do
   -- lift $ traceparse "numericquantityp"
-  (q, _, _, _) <- lift $ numberp Nothing
+  (q, _, _, _) <- numberp Nothing
   msymbol <- optional $ choice $ map (string . fst) timeUnits
-  lift skipNonNewlineSpaces
+  skipNonNewlineSpaces
   let q' =
         case msymbol of
           Nothing  -> q
@@ -239,7 +249,7 @@ timeUnits =
 -- @
 -- .... ..
 -- @
-dotquantityp :: JournalParser m Quantity
+dotquantityp :: TextParser m Quantity
 dotquantityp = do
   -- lift $ traceparse "dotquantityp"
   dots <- filter (not.isSpace) <$> many (oneOf (". " :: [Char]))
