@@ -26,9 +26,9 @@ import Data.Decimal (roundTo)
 import Data.Function (on)
 import Data.HashMap.Strict (HashMap)
 import qualified Data.HashMap.Strict as HM
-import Data.List (find, partition, transpose, foldl')
+import Data.List (find, partition, transpose, foldl', maximumBy)
 import Data.List.Extra (nubSort)
-import Data.Maybe (fromMaybe, catMaybes)
+import Data.Maybe (fromMaybe, catMaybes, isJust)
 import Data.Map (Map)
 import qualified Data.Map as Map
 import qualified Data.Set as S
@@ -36,6 +36,7 @@ import Data.Text (Text)
 import qualified Data.Text as T
 import qualified Data.Text.Lazy as TL
 import qualified Data.Text.Lazy.Builder as TB
+import Safe (minimumDef)
 --import System.Console.CmdArgs.Explicit as C
 --import Lucid as L
 import qualified Text.Tabular.AsciiWide as Tab
@@ -45,6 +46,8 @@ import Hledger.Utils
 import Hledger.Reports.ReportOptions
 import Hledger.Reports.ReportTypes
 import Hledger.Reports.MultiBalanceReport
+import Data.Ord (comparing)
+import Control.Monad ((>=>))
 
 
 type BudgetGoal    = Change
@@ -78,7 +81,7 @@ budgetReport rspec bopts reportspan j = dbg4 "sortedbudgetreport" budgetreport
       expandAccountNames $
       accountNamesFromPostings $
       concatMap tpostings $
-      concatMap (`runPeriodicTransaction` reportspan) $
+      concatMap (\pt -> runPeriodicTransaction False pt reportspan) $
       jperiodictxns j
     actualj = journalWithBudgetAccountNames budgetedaccts showunbudgeted j
     budgetj = journalAddBudgetGoalTransactions bopts ropts reportspan j
@@ -96,28 +99,64 @@ budgetReport rspec bopts reportspan j = dbg4 "sortedbudgetreport" budgetreport
       | otherwise = budgetgoalreport
     budgetreport = combineBudgetAndActual ropts j budgetgoalreport' actualreport
 
--- | Use all periodic transactions in the journal to generate
--- budget goal transactions in the specified date span.
--- Budget goal transactions are similar to forecast transactions except
--- their purpose and effect is to define balance change goals, per account and period,
--- for BudgetReport.
+-- | Use all (or all matched by --budget's argument) periodic transactions in the journal 
+-- to generate budget goal transactions in the specified date span (and before, to support
+-- --historical. The precise start date is the natural start date of the largest interval
+-- of the active periodic transaction rules that is on or before the earlier of journal start date,
+-- report start date.)
+-- Budget goal transactions are similar to forecast transactions except their purpose 
+-- and effect is to define balance change goals, per account and period, for BudgetReport.
+--
 journalAddBudgetGoalTransactions :: BalancingOpts -> ReportOpts -> DateSpan -> Journal -> Journal
 journalAddBudgetGoalTransactions bopts ropts reportspan j =
-  either error' id $ journalBalanceTransactions bopts j{ jtxns = budgetts }  -- PARTIAL:
+  either error' id $  -- PARTIAL:
+    (journalApplyCommodityStyles >=> journalBalanceTransactions bopts) j{ jtxns = budgetts }
   where
-    budgetspan = dbg3 "budget span" $ reportspan
-    pat = fromMaybe "" $ dbg3 "budget pattern" $ T.toLower <$> budgetpat_ ropts
+    budgetspan = dbg3 "budget span" $ DateSpan (Exact <$> mbudgetgoalsstartdate) (Exact <$> spanEnd reportspan)
+      where
+        mbudgetgoalsstartdate =
+          -- We want to also generate budget goal txns before the report start date, in case -H is used.
+          -- What should the actual starting date for goal txns be ? This gets tricky. 
+          -- Consider a journal with a "~ monthly" periodic transaction rule, where the first transaction is on 1/5.
+          -- Users will certainly expect a budget goal for january, but "~ monthly" generates transactions
+          -- on the first of month, and starting from 1/5 would exclude 1/1.
+          -- Secondly, consider a rule like "~ every february 2nd from 2020/01"; we should not start that
+          -- before 2020-02-02.
+          -- Hopefully the following algorithm produces intuitive behaviour in general:
+          -- from the earlier of the journal start date and the report start date,
+          -- move backward to the nearest natural start date of the largest period seen among the
+          -- active periodic transactions, unless that is disallowed by a start date in the periodic rule.
+          -- (Do we need to pay attention to an end date in the rule ? Don't think so.)
+          -- (So with "~ monthly", the journal start date 1/5 is adjusted to 1/1.)
+          case minimumDef Nothing $ filter isJust [journalStartDate False j, spanStart reportspan] of
+            Nothing -> Nothing
+            Just d  -> Just d'
+              where
+                -- the interval and any date span of the periodic transaction with longest period
+                (intervl, spn) =
+                  case budgetpts of
+                    []  -> (Days 1, nulldatespan)
+                    pts -> (ptinterval pt, ptspan pt)
+                      where pt = maximumBy (comparing ptinterval) pts  -- PARTIAL: maximumBy won't fail
+                -- the natural start of this interval on or before the journal/report start
+                intervalstart = intervalBoundaryBefore intervl d
+                -- the natural interval start before the journal/report start,
+                -- or the rule-specified start if later,
+                -- but no later than the journal/report start.
+                d' = min d $ maybe intervalstart (max intervalstart) $ spanStart spn
+
     -- select periodic transactions matching a pattern
     -- (the argument of the (final) --budget option).
     -- XXX two limitations/wishes, requiring more extensive type changes:
     -- - give an error if pat is non-null and matches no periodic txns
     -- - allow a regexp or a full hledger query, not just a substring
+    pat = fromMaybe "" $ dbg3 "budget pattern" $ T.toLower <$> budgetpat_ ropts
+    budgetpts = [pt | pt <- jperiodictxns j, pat `T.isInfixOf` T.toLower (ptdescription pt)]
     budgetts =
       dbg5 "budget goal txns" $
       [makeBudgetTxn t
-      | pt <- jperiodictxns j
-      , pat `T.isInfixOf` T.toLower (ptdescription pt)
-      , t <- runPeriodicTransaction pt budgetspan
+      | pt <- budgetpts
+      , t <- runPeriodicTransaction False pt budgetspan
       ]
     makeBudgetTxn t = txnTieKnot $ t { tdescription = T.pack "Budget transaction" }
 
@@ -269,11 +308,11 @@ budgetReportAsTable
       | transpose_ = \(Tab.Table rh ch vals) -> Tab.Table ch rh (transpose vals)
       | otherwise  = id
 
-    (accts, rows, totalrows) = (accts, prependcs itemscs (padcells texts), prependcs trcs (padtr trtexts))
+    (accts, rows, totalrows) = (accts', prependcs itemscs (padcells texts), prependcs trcs (padtr trtexts))
       where
         shownitems :: [[(AccountName, WideBuilder, BudgetDisplayRow)]]
         shownitems = (fmap (\i -> fmap (\(cs, cvals) -> (renderacct i, cs, cvals)) . showrow $ rowToBudgetCells i) items)
-        (accts, itemscs, texts) = unzip3 $ concat shownitems
+        (accts', itemscs, texts) = unzip3 $ concat shownitems
 
         showntr    :: [[(WideBuilder, BudgetDisplayRow)]]
         showntr    = [showrow $ rowToBudgetCells tr]
@@ -342,10 +381,8 @@ budgetReportAsTable
       where
         actual' = fromMaybe nullmixedamt actual
 
-        budgetAndPerc b = uncurry zip
-          ( showmixed b
-          , fmap (wbFromText . T.pack . show . roundTo 0) <$> percbudget actual' b
-          )
+        budgetAndPerc b = 
+          zip (showmixed b) (fmap (wbFromText . T.pack . show . roundTo 0) <$> percbudget actual' b)
 
         full
           | Just b <- mbudget = Just <$> budgetAndPerc b
@@ -358,9 +395,9 @@ budgetReportAsTable
             (TB.fromText . flip T.replicate " " $ actualwidth - w) <> b
 
         (totalpercentwidth, totalbudgetwidth) =
-          let totalpercentwidth = if percentwidth == 0 then 0 else percentwidth + 5
-           in ( totalpercentwidth
-              , if budgetwidth == 0 then 0 else budgetwidth + totalpercentwidth + 3
+          let totalpercentwidth' = if percentwidth == 0 then 0 else percentwidth + 5
+           in ( totalpercentwidth'
+              , if budgetwidth == 0 then 0 else budgetwidth + totalpercentwidth' + 3
               )
 
         -- | Display a padded budget string
@@ -407,14 +444,20 @@ budgetReportAsCsv
   (PeriodicReport colspans items tr)
   = (if transpose_ then transpose else id) $
 
+  -- heading row
+  
+
   -- heading row
   ("Account" :
   ["Commodity" | layout_ == LayoutBare ]
-   ++ concatMap (\span -> [showDateSpan span, "budget"]) colspans
+   ++ concatMap (\spn -> [showDateSpan spn, "budget"]) colspans
    ++ concat [["Total"  ,"budget"] | row_total_]
    ++ concat [["Average","budget"] | average_]
   ) :
 
+  -- account rows
+  
+
   -- account rows
   concatMap (rowAsTexts prrFullName) items
 
@@ -422,23 +465,23 @@ budgetReportAsCsv
   ++ concat [ rowAsTexts (const "Total:") tr | not no_total_ ]
 
   where
-    flattentuples abs = concat [[a,b] | (a,b) <- abs]
+    flattentuples tups = concat [[a,b] | (a,b) <- tups]
     showNorm = maybe "" (wbToText . showMixedAmountB oneLine)
 
     rowAsTexts :: (PeriodicReportRow a BudgetCell -> Text)
                -> PeriodicReportRow a BudgetCell
                -> [[Text]]
     rowAsTexts render row@(PeriodicReportRow _ as (rowtot,budgettot) (rowavg, budgetavg))
-      | layout_ /= LayoutBare = [render row : fmap showNorm all]
+      | layout_ /= LayoutBare = [render row : fmap showNorm vals]
       | otherwise =
             joinNames . zipWith (:) cs  -- add symbols and names
           . transpose                   -- each row becomes a list of Text quantities
           . fmap (fmap wbToText . showMixedAmountLinesB oneLine{displayOrder=Just cs, displayMinWidth=Nothing}
                  .fromMaybe nullmixedamt)
-          $ all
+          $ vals
       where
-        cs = S.toList . foldl' S.union mempty . fmap maCommodities $ catMaybes all
-        all = flattentuples as
+        cs = S.toList . foldl' S.union mempty . fmap maCommodities $ catMaybes vals
+        vals = flattentuples as
             ++ concat [[rowtot, budgettot] | row_total_]
             ++ concat [[rowavg, budgetavg] | average_]
 
