@@ -165,15 +165,14 @@ compoundBalanceReportWith rspec' j priceoracle subreportspecs = cbr
             , cbcsubreportincreasestotal
             )
           where
+            ropts = cbcsubreportoptions $ _rsReportOpts rspec
             -- Add a restriction to this subreport to the report query.
             -- XXX in non-thorough way, consider updateReportSpec ?
-            q = cbcsubreportquery j
-            ropts = cbcsubreportoptions $ _rsReportOpts rspec
-            rspecsub = rspec{_rsReportOpts=ropts, _rsQuery=And [q, _rsQuery rspec]}
+            rspecsub = rspec{_rsReportOpts=ropts, _rsQuery=And [cbcsubreportquery, _rsQuery rspec]}
             -- Starting balances and column postings specific to this subreport.
             startbals' = startingBalances rspecsub j priceoracle $
-              filter (matchesPostingExtra (journalAccountType j) q) startps
-            colps' = map (second $ filter (matchesPostingExtra (journalAccountType j) q)) colps
+              filter (matchesPostingExtra (journalAccountType j) cbcsubreportquery) startps
+            colps' = map (second $ filter (matchesPostingExtra (journalAccountType j) cbcsubreportquery)) colps
 
     -- Sum the subreport totals by column. Handle these cases:
     -- - no subreports
@@ -223,7 +222,7 @@ startingPostings rspec@ReportSpec{_rsQuery=query,_rsReportOpts=ropts} j priceora
 
     precedingperiod = dateSpanAsPeriod . spanIntersect precedingspan .
                          periodAsDateSpan $ period_ ropts
-    precedingspan = DateSpan Nothing $ spanStart reportspan
+    precedingspan = DateSpan Nothing (Exact <$> spanStart reportspan)
     precedingspanq = (if date2_ ropts then Date2 else Date) $ case precedingspan of
         DateSpan Nothing Nothing -> emptydatespan
         a -> a
@@ -315,22 +314,25 @@ calculateReportMatrix rspec@ReportSpec{_rsReportOpts=ropts} j priceoracle startb
     -- The valued row amounts to be displayed: per-period changes,
     -- zero-based cumulative totals, or
     -- starting-balance-based historical balances.
-    rowbals name changes = dbg5 "rowbals" $ case balanceaccum_ ropts of
-        PerPeriod  -> changeamts
+    rowbals name unvaluedChanges = dbg5 "rowbals" $ case balanceaccum_ ropts of
+        PerPeriod  -> changes
         Cumulative -> cumulative
         Historical -> historical
       where
-        -- changes to report on: usually just the changes itself, but use the
-        -- differences in the historical amount for ValueChangeReports.
-        changeamts = case balancecalc_ ropts of
-            CalcChange      -> M.mapWithKey avalue changes
-            CalcBudget      -> M.mapWithKey avalue changes
-            CalcValueChange -> periodChanges valuedStart historical
-            CalcGain        -> periodChanges valuedStart historical
-        cumulative = cumulativeSum avalue nullacct changeamts
-        historical = cumulativeSum avalue startingBalance changes
+        -- changes to report on: usually just the valued changes themselves, but use the
+        -- differences in the valued historical amount for CalcValueChange and CalcGain.
+        changes = case balancecalc_ ropts of
+            CalcChange        -> M.mapWithKey avalue unvaluedChanges
+            CalcBudget        -> M.mapWithKey avalue unvaluedChanges
+            CalcValueChange   -> periodChanges valuedStart historical
+            CalcGain          -> periodChanges valuedStart historical
+            CalcPostingsCount -> M.mapWithKey avalue unvaluedChanges
+        -- the historical balance is the valued cumulative sum of all unvalued changes
+        historical = M.mapWithKey avalue $ cumulativeSum startingBalance unvaluedChanges
+        -- since this is a cumulative sum of valued amounts, it should not be valued again
+        cumulative = cumulativeSum nullacct changes
         startingBalance = HM.lookupDefault nullacct name startbals
-        valuedStart = avalue (DateSpan Nothing historicalDate) startingBalance
+        valuedStart = avalue (DateSpan Nothing (Exact <$> historicalDate)) startingBalance
 
     -- In each column, get each account's balance changes
     colacctchanges = dbg5 "colacctchanges" $ map (second $ acctChanges rspec j) colps :: [(DateSpan, HashMap ClippedAccountName Account)]
@@ -342,7 +344,7 @@ calculateReportMatrix rspec@ReportSpec{_rsReportOpts=ropts} j priceoracle startb
     avalue = acctApplyBoth . mixedAmountApplyValuationAfterSumFromOptsWith ropts j priceoracle
     acctApplyBoth f a = a{aibalance = f $ aibalance a, aebalance = f $ aebalance a}
     historicalDate = minimumMay $ mapMaybe spanStart colspans
-    zeros = M.fromList [(span, nullacct) | span <- colspans]
+    zeros = M.fromList [(spn, nullacct) | spn <- colspans]
     colspans = map fst colps
 
 
@@ -352,9 +354,15 @@ calculateReportMatrix rspec@ReportSpec{_rsReportOpts=ropts} j priceoracle startb
 generateMultiBalanceReport :: ReportSpec -> Journal -> PriceOracle -> Set AccountName
                            -> [(DateSpan, [Posting])] -> HashMap AccountName Account
                            -> MultiBalanceReport
-generateMultiBalanceReport rspec@ReportSpec{_rsReportOpts=ropts} j priceoracle unelidableaccts colps startbals =
+generateMultiBalanceReport rspec@ReportSpec{_rsReportOpts=ropts} j priceoracle unelidableaccts colps0 startbals =
     report
   where
+    -- If doing --count, set all posting amounts to "1".
+    colps =
+      if balancecalc_ ropts == CalcPostingsCount
+      then map (second (map (postingTransformAmount (const $ mixed [num 1])))) colps0
+      else colps0
+
     -- Process changes into normal, cumulative, or historical amounts, plus value them
     matrix = calculateReportMatrix rspec j priceoracle startbals colps
 
@@ -405,11 +413,11 @@ displayedAccounts :: ReportSpec
                   -> HashMap AccountName (Map DateSpan Account)
                   -> HashMap AccountName DisplayName
 displayedAccounts ReportSpec{_rsQuery=query,_rsReportOpts=ropts} unelidableaccts valuedaccts
-    | depth == 0 = HM.singleton "..." $ DisplayName "..." "..." 1
+    | qdepth == 0 = HM.singleton "..." $ DisplayName "..." "..." 1
     | otherwise  = HM.mapWithKey (\a _ -> displayedName a) displayedAccts
   where
     -- Accounts which are to be displayed
-    displayedAccts = (if depth == 0 then id else HM.filterWithKey keep) valuedaccts
+    displayedAccts = (if qdepth == 0 then id else HM.filterWithKey keep) valuedaccts
       where
         keep name amts = isInteresting name amts || name `HM.member` interestingParents
 
@@ -428,7 +436,7 @@ displayedAccounts ReportSpec{_rsQuery=query,_rsReportOpts=ropts} unelidableaccts
 
     -- Accounts interesting for their own sake
     isInteresting name amts =
-        d <= depth                                 -- Throw out anything too deep
+        d <= qdepth                                  -- Throw out anything too deep
         && ( name `Set.member` unelidableaccts     -- Unelidable accounts should be kept unless too deep
            ||(empty_ ropts && keepWhenEmpty amts)  -- Keep empty accounts when called with --empty
            || not (isZeroRow balance amts)         -- Keep everything with a non-zero balance in the row
@@ -439,8 +447,8 @@ displayedAccounts ReportSpec{_rsQuery=query,_rsReportOpts=ropts} unelidableaccts
             ALFlat -> const True          -- Keep all empty accounts in flat mode
             ALTree -> all (null . asubs)  -- Keep only empty leaves in tree mode
         balance = maybeStripPrices . case accountlistmode_ ropts of
-            ALTree | d == depth -> aibalance
-            _                   -> aebalance
+            ALTree | d == qdepth -> aibalance
+            _                    -> aebalance
           where maybeStripPrices = if conversionop_ ropts == Just NoConversionOp then id else mixedAmountStripPrices
 
     -- Accounts interesting because they are a fork for interesting subaccounts
@@ -452,7 +460,7 @@ displayedAccounts ReportSpec{_rsQuery=query,_rsReportOpts=ropts} unelidableaccts
         minSubs = if no_elide_ ropts then 1 else 2
 
     isZeroRow balance = all (mixedAmountLooksZero . balance)
-    depth = fromMaybe maxBound $  queryDepth query
+    qdepth = fromMaybe maxBound $ queryDepth query
     numSubs = subaccountTallies . HM.keys $ HM.filterWithKey isInteresting valuedaccts
 
 -- | Sort the rows by amount or by account declaration order.
@@ -533,10 +541,10 @@ transposeMap :: [(DateSpan, HashMap AccountName a)]
              -> HashMap AccountName (Map DateSpan a)
 transposeMap = foldr (uncurry addSpan) mempty
   where
-    addSpan span acctmap seen = HM.foldrWithKey (addAcctSpan span) seen acctmap
+    addSpan spn acctmap seen = HM.foldrWithKey (addAcctSpan spn) seen acctmap
 
-    addAcctSpan span acct a = HM.alter f acct
-      where f = Just . M.insert span a . fromMaybe mempty
+    addAcctSpan spn acct a = HM.alter f acct
+      where f = Just . M.insert spn a . fromMaybe mempty
 
 -- | A sorting helper: sort a list of things (eg report rows) keyed by account name
 -- to match the provided ordering of those same account names.
@@ -580,11 +588,9 @@ periodChanges start amtmap =
     M.fromDistinctAscList . zip dates $ zipWith subtractAcct amts (start:amts)
   where (dates, amts) = unzip $ M.toAscList amtmap
 
--- | Calculate a cumulative sum from a list of period changes and a valuation
--- function.
-cumulativeSum :: (DateSpan -> Account -> Account) -> Account -> Map DateSpan Account -> Map DateSpan Account
-cumulativeSum value start = snd . M.mapAccumWithKey accumValued start
-  where accumValued startAmt date newAmt = let s = sumAcct startAmt newAmt in (s, value date s)
+-- | Calculate a cumulative sum from a list of period changes.
+cumulativeSum :: Account -> Map DateSpan Account -> Map DateSpan Account
+cumulativeSum start = snd . M.mapAccum (\a b -> let s = sumAcct a b in (s, s)) start
 
 -- | Given a table representing a multi-column balance report (for example,
 -- made using 'balanceReportAsTable'), render it in a format suitable for
