@@ -23,7 +23,7 @@ where
 
 import Data.List (nub, sortOn)
 import Data.List.Extra (nubSort)
-import Data.Maybe (fromMaybe, isJust, isNothing)
+import Data.Maybe (isJust, isNothing)
 import Data.Text (Text)
 import Data.Time.Calendar (Day)
 import Safe (headMay)
@@ -73,8 +73,8 @@ postingsReport rspec@ReportSpec{_rsReportOpts=ropts@ReportOpts{..}} j = items
 
       -- Postings, or summary postings with their subperiod's end date, to be displayed.
       displayps :: [(Posting, Maybe Period)]
-        | multiperiod = [(p, Just period) | (p, period) <- summariseps reportps]
-        | otherwise   = [(p, Nothing) | p <- reportps]
+        | multiperiod = [(p', Just period') | (p', period') <- summariseps reportps]
+        | otherwise   = [(p', Nothing) | p' <- reportps]
         where
           summariseps = summarisePostingsByInterval whichdate mdepth showempty colspans
           showempty = empty_ || average_
@@ -115,11 +115,16 @@ matchedPostingsBeforeAndDuring :: ReportSpec -> Journal -> DateSpan -> ([Posting
 matchedPostingsBeforeAndDuring rspec@ReportSpec{_rsReportOpts=ropts,_rsQuery=q} j reportspan =
     dbg5 "beforeps, duringps" $ span (beforestartq `matchesPosting`) beforeandduringps
   where
-    beforestartq = dbg3 "beforestartq" $ dateqtype $ DateSpan Nothing $ spanStart reportspan
+    beforestartq = dbg3 "beforestartq" $ dateqtype $ DateSpan Nothing (Exact <$> spanStart reportspan)
     beforeandduringps = 
         sortOn (postingDateOrDate2 (whichDate ropts))            -- sort postings by date or date2
       . (if invert_ ropts then map negatePostingAmount else id)  -- with --invert, invert amounts
       . journalPostings
+      -- With most calls we will not require transaction prices past this point, and can get a big
+      -- speed improvement by stripping them early. In some cases, such as in hledger-ui, we still
+      -- want to keep prices around, so we can toggle between cost and no cost quickly. We can use
+      -- the show_costs_ flag to be efficient when we can, and detailed when we have to.
+      . (if show_costs_ ropts then id else journalMapPostingAmounts mixedAmountStripPrices)
       $ journalValueAndFilterPostings rspec{_rsQuery=beforeandduringq} j
 
     -- filter postings by the query, with no start date or depth limit
@@ -127,7 +132,7 @@ matchedPostingsBeforeAndDuring rspec@ReportSpec{_rsReportOpts=ropts,_rsQuery=q}
       where
         depthless  = filterQuery (not . queryIsDepth)
         dateless   = filterQuery (not . queryIsDateOrDate2)
-        beforeendq = dateqtype $ DateSpan Nothing $ spanEnd reportspan
+        beforeendq = dateqtype $ DateSpan Nothing (Exact <$> spanEnd reportspan)
 
     dateqtype = if queryIsDate2 dateq || (queryIsDate dateq && date2_ ropts) then Date2 else Date
       where
@@ -184,25 +189,25 @@ summarisePostingsByInterval wd mdepth showempty colspans =
 -- with 0 amount.
 --
 summarisePostingsInDateSpan :: DateSpan -> WhichDate -> Maybe Int -> Bool -> [Posting] -> [SummaryPosting]
-summarisePostingsInDateSpan span@(DateSpan b e) wd mdepth showempty ps
+summarisePostingsInDateSpan spn@(DateSpan b e) wd mdepth showempty ps
   | null ps && (isNothing b || isNothing e) = []
-  | null ps && showempty = [(summaryp, dateSpanAsPeriod span)]
+  | null ps && showempty = [(summaryp, dateSpanAsPeriod spn)]
   | otherwise = summarypes
   where
     postingdate = if wd == PrimaryDate then postingDate else postingDate2
-    b' = fromMaybe (maybe nulldate postingdate $ headMay ps) b
+    b' = maybe (maybe nulldate postingdate $ headMay ps) fromEFDay b
     summaryp = nullposting{pdate=Just b'}
     clippedanames = nub $ map (clipAccountName mdepth) anames
     summaryps | mdepth == Just 0 = [summaryp{paccount="...",pamount=sumPostings ps}]
               | otherwise        = [summaryp{paccount=a,pamount=balance a} | a <- clippedanames]
-    summarypes = map (, dateSpanAsPeriod span) $ (if showempty then id else filter (not . mixedAmountLooksZero . pamount)) summaryps
+    summarypes = map (, dateSpanAsPeriod spn) $ (if showempty then id else filter (not . mixedAmountLooksZero . pamount)) summaryps
     anames = nubSort $ map paccount ps
     -- aggregate balances by account, like ledgerFromJournal, then do depth-clipping
     accts = accountsFromPostings ps
     balance a = maybe nullmixedamt bal $ lookupAccount a accts
       where
         bal = if isclipped a then aibalance else aebalance
-        isclipped a = maybe False (accountNameLevel a >=) mdepth
+        isclipped a' = maybe False (accountNameLevel a' >=) mdepth
 
 negatePostingAmount :: Posting -> Posting
 negatePostingAmount = postingTransformAmount negate
