@@ -13,6 +13,7 @@ Options common to most hledger reports.
 {-# LANGUAGE RecordWildCards       #-}
 {-# LANGUAGE TemplateHaskell       #-}
 {-# LANGUAGE TypeFamilies          #-}
+{-# LANGUAGE TypeOperators #-}
 
 module Hledger.Reports.ReportOptions (
   ReportOpts(..),
@@ -63,14 +64,15 @@ module Hledger.Reports.ReportOptions (
 )
 where
 
-import Control.Applicative (Const(..), (<|>), liftA2)
+import Prelude hiding (Applicative(..))
+import Control.Applicative (Applicative(..), Const(..), (<|>))
 import Control.Monad ((<=<), guard, join)
 import Data.Char (toLower)
 import Data.Either (fromRight)
 import Data.Either.Extra (eitherToMaybe)
 import Data.Functor.Identity (Identity(..))
 import Data.List.Extra (find, isPrefixOf, nubSort)
-import Data.Maybe (fromMaybe, isJust)
+import Data.Maybe (fromMaybe, isJust, isNothing)
 import qualified Data.Text as T
 import Data.Time.Calendar (Day, addDays)
 import Data.Default (Default(..))
@@ -86,10 +88,11 @@ import Hledger.Utils
 -- | What to calculate for each cell in a balance report.
 -- "Balance report types -> Calculation type" in the hledger manual.
 data BalanceCalculation =
-    CalcChange      -- ^ Sum of posting amounts in the period.
-  | CalcBudget      -- ^ Sum of posting amounts and the goal for the period.
-  | CalcValueChange -- ^ Change from previous period's historical end value to this period's historical end value.
-  | CalcGain        -- ^ Change from previous period's gain, i.e. valuation minus cost basis.
+    CalcChange        -- ^ Sum of posting amounts in the period.
+  | CalcBudget        -- ^ Sum of posting amounts and the goal for the period.
+  | CalcValueChange   -- ^ Change from previous period's historical end value to this period's historical end value.
+  | CalcGain          -- ^ Change from previous period's gain, i.e. valuation minus cost basis.
+  | CalcPostingsCount -- ^ Number of postings in the period.
   deriving (Eq, Show)
 
 instance Default BalanceCalculation where def = CalcChange
@@ -266,13 +269,14 @@ rawOptsToReportOpts d rawopts =
           ,layout_           = layoutopt rawopts
           }
 
--- | The result of successfully parsing a ReportOpts on a particular
--- Day. Any ambiguous dates are completed and Queries are parsed,
--- ensuring that there are no regular expression errors. Values here
--- should be used in preference to re-deriving them from ReportOpts.
--- If you change the query_ in ReportOpts, you should call
--- `reportOptsToSpec` to regenerate the ReportSpec with the new
--- Query.
+-- | A fully-determined set of report parameters 
+-- (report options with all partial values made total, eg the begin and end
+-- dates are known, avoiding date/regex errors; plus the reporting date),
+-- and the query successfully calculated from them.
+--
+-- If you change the report options or date in one of these, you should
+-- use `reportOptsToSpec` to regenerate the whole thing, avoiding inconsistency.
+--
 data ReportSpec = ReportSpec
   { _rsReportOpts :: ReportOpts  -- ^ The underlying ReportOpts used to generate this ReportSpec
   , _rsDay        :: Day         -- ^ The Day this ReportSpec is generated for
@@ -292,8 +296,8 @@ defreportspec = ReportSpec
 
 -- | Set the default ConversionOp.
 setDefaultConversionOp :: ConversionOp -> ReportSpec -> ReportSpec
-setDefaultConversionOp def rspec@ReportSpec{_rsReportOpts=ropts} =
-    rspec{_rsReportOpts=ropts{conversionop_=conversionop_ ropts <|> Just def}}
+setDefaultConversionOp defop rspec@ReportSpec{_rsReportOpts=ropts} =
+    rspec{_rsReportOpts=ropts{conversionop_=conversionop_ ropts <|> Just defop}}
 
 accountlistmodeopt :: RawOpts -> AccountListMode
 accountlistmodeopt =
@@ -315,6 +319,7 @@ balancecalcopt =
       "valuechange" -> Just CalcValueChange
       "gain"        -> Just CalcGain
       "budget"      -> Just CalcBudget
+      "count"       -> Just CalcPostingsCount
       _             -> Nothing
 
 balanceaccumopt :: RawOpts -> BalanceAccumulation
@@ -360,7 +365,7 @@ layoutopt rawopts = fromMaybe (LayoutWide Nothing) $ layout <|> column
         (s,n) = break (==',') $ map toLower opt
         w = case drop 1 n of
               "" -> Nothing
-              c | Just w <- readMay c -> Just w
+              c | Just w' <- readMay c -> Just w'
               _ -> usageError "width in --layout=wide,WIDTH must be an integer"
 
         err = usageError "--layout's argument should be \"wide[,WIDTH]\", \"tall\", \"bare\", or \"tidy\""
@@ -380,24 +385,24 @@ periodFromRawOpts d rawopts =
   where
     mlastb = case beginDatesFromRawOpts d rawopts of
                    [] -> Nothing
-                   bs -> Just $ last bs
+                   bs -> Just $ fromEFDay $ last bs
     mlaste = case endDatesFromRawOpts d rawopts of
                    [] -> Nothing
-                   es -> Just $ last es
+                   es -> Just $ fromEFDay $ last es
 
 -- Get all begin dates specified by -b/--begin or -p/--period options, in order,
 -- using the given date to interpret relative date expressions.
-beginDatesFromRawOpts :: Day -> RawOpts -> [Day]
+beginDatesFromRawOpts :: Day -> RawOpts -> [EFDay]
 beginDatesFromRawOpts d = collectopts (begindatefromrawopt d)
   where
-    begindatefromrawopt d (n,v)
+    begindatefromrawopt d' (n,v)
       | n == "begin" =
           either (\e -> usageError $ "could not parse "++n++" date: "++customErrorBundlePretty e) Just $
-          fixSmartDateStrEither' d (T.pack v)
+          fixSmartDateStrEither' d' (T.pack v)
       | n == "period" =
         case
           either (\e -> usageError $ "could not parse period option: "++customErrorBundlePretty e) id $
-          parsePeriodExpr d (stripquotes $ T.pack v)
+          parsePeriodExpr d' (stripquotes $ T.pack v)
         of
           (_, DateSpan (Just b) _) -> Just b
           _                        -> Nothing
@@ -405,17 +410,17 @@ beginDatesFromRawOpts d = collectopts (begindatefromrawopt d)
 
 -- Get all end dates specified by -e/--end or -p/--period options, in order,
 -- using the given date to interpret relative date expressions.
-endDatesFromRawOpts :: Day -> RawOpts -> [Day]
+endDatesFromRawOpts :: Day -> RawOpts -> [EFDay]
 endDatesFromRawOpts d = collectopts (enddatefromrawopt d)
   where
-    enddatefromrawopt d (n,v)
+    enddatefromrawopt d' (n,v)
       | n == "end" =
           either (\e -> usageError $ "could not parse "++n++" date: "++customErrorBundlePretty e) Just $
-          fixSmartDateStrEither' d (T.pack v)
+          fixSmartDateStrEither' d' (T.pack v)
       | n == "period" =
         case
           either (\e -> usageError $ "could not parse period option: "++customErrorBundlePretty e) id $
-          parsePeriodExpr d (stripquotes $ T.pack v)
+          parsePeriodExpr d' (stripquotes $ T.pack v)
         of
           (_, DateSpan _ (Just e)) -> Just e
           _                        -> Nothing
@@ -589,16 +594,16 @@ journalApplyValuationFromOptsWith rspec@ReportSpec{_rsReportOpts=ropts} j priceo
       CalcGain -> journalMapPostings (\p -> postingTransformAmount (gain p) p) j
       _        -> journalMapPostings (\p -> postingTransformAmount (valuation p) p) $ costing j
   where
-    valuation p = maybe id (mixedAmountApplyValuation priceoracle styles (periodEnd p) (_rsDay rspec) (postingDate p)) (value_ ropts)
-    gain      p = maybe id (mixedAmountApplyGain      priceoracle styles (periodEnd p) (_rsDay rspec) (postingDate p)) (value_ ropts)
+    valuation p = maybe id (mixedAmountApplyValuation priceoracle styles (postingperiodend p) (_rsDay rspec) (postingDate p)) (value_ ropts)
+    gain      p = maybe id (mixedAmountApplyGain      priceoracle styles (postingperiodend p) (_rsDay rspec) (postingDate p)) (value_ ropts)
     costing     = journalToCost (fromMaybe NoConversionOp $ conversionop_ ropts)
 
     -- Find the end of the period containing this posting
-    periodEnd  = addDays (-1) . fromMaybe err . mPeriodEnd . postingDate
+    postingperiodend  = addDays (-1) . fromMaybe err . mPeriodEnd . postingDateOrDate2 (whichDate ropts)
     mPeriodEnd = case interval_ ropts of
         NoInterval -> const . spanEnd . fst $ reportSpan j rspec
         _          -> spanEnd <=< latestSpanContaining (historical : spans)
-    historical = DateSpan Nothing $ spanStart =<< headMay spans
+    historical = DateSpan Nothing $ (fmap Exact . spanStart) =<< headMay spans
     spans = snd $ reportSpanBothDates j rspec
     styles = journalCommodityStyles j
     err = error "journalApplyValuationFromOpts: expected all spans to have an end date"
@@ -611,11 +616,11 @@ mixedAmountApplyValuationAfterSumFromOptsWith ropts j priceoracle =
     case valuationAfterSum ropts of
         Just mc -> case balancecalc_ ropts of
             CalcGain -> gain mc
-            _        -> \span -> valuation mc span . costing
+            _        -> \spn -> valuation mc spn . costing
         Nothing      -> const id
   where
-    valuation mc span = mixedAmountValueAtDate priceoracle styles mc (maybe err (addDays (-1)) $ spanEnd span)
-    gain mc span = mixedAmountGainAtDate priceoracle styles mc (maybe err (addDays (-1)) $ spanEnd span)
+    valuation mc spn = mixedAmountValueAtDate priceoracle styles mc (maybe err (addDays (-1)) $ spanEnd spn)
+    gain mc spn = mixedAmountGainAtDate priceoracle styles mc (maybe err (addDays (-1)) $ spanEnd spn)
     costing = case fromMaybe NoConversionOp $ conversionop_ ropts of
         NoConversionOp -> id
         ToCost         -> styleMixedAmount styles . mixedAmountCost
@@ -674,7 +679,7 @@ reportSpanHelper bothdates j ReportSpec{_rsQuery=query, _rsReportOpts=ropts} =
     -- include price directives after the last transaction
     journalspan = dbg3 "journalspan" $ if bothdates then journalDateSpanBothDates j else journalDateSpan (date2_ ropts) j
     pricespan = dbg3 "pricespan" . DateSpan Nothing $ case value_ ropts of
-        Just (AtEnd _) -> fmap (addDays 1) . maximumMay . map pddate $ jpricedirectives j
+        Just (AtEnd _) -> fmap (Exact . addDays 1) . maximumMay . map pddate $ jpricedirectives j
         _              -> Nothing
     -- If the requested span is open-ended, close it using the journal's start and end dates.
     -- This can still be the null (open) span if the journal is empty.
@@ -683,11 +688,15 @@ reportSpanHelper bothdates j ReportSpec{_rsQuery=query, _rsReportOpts=ropts} =
     -- This list can be empty if the journal was empty,
     -- or if hledger-ui has added its special date:-tomorrow to the query
     -- and all txns are in the future.
-    intervalspans  = dbg3 "intervalspans" $ splitSpan (interval_ ropts) requestedspan'
+    intervalspans  = dbg3 "intervalspans" $ splitSpan adjust (interval_ ropts) requestedspan'
+      where
+        -- When calculating report periods, we will adjust the start date back to the nearest interval boundary
+        -- unless a start date was specified explicitly.
+        adjust = isNothing $ spanStart requestedspan
     -- The requested span enlarged to enclose a whole number of intervals.
     -- This can be the null span if there were no intervals.
-    reportspan = dbg3 "reportspan" $ DateSpan (spanStart =<< headMay intervalspans)
-                                              (spanEnd =<< lastMay intervalspans)
+    reportspan = dbg3 "reportspan" $ DateSpan (fmap Exact . spanStart =<< headMay intervalspans)
+                                              (fmap Exact . spanEnd =<< lastMay intervalspans)
 
 reportStartDate :: Journal -> ReportSpec -> Maybe Day
 reportStartDate j = spanStart . fst . reportSpan j
@@ -797,11 +806,11 @@ makeHledgerClassyLenses ''ReportSpec
 -- >>> _rsQuery <$> setEither querystring ["assets"] defreportspec
 -- Right (Acct (RegexpCI "assets"))
 -- >>> _rsQuery <$> setEither querystring ["(assets"] defreportspec
--- Left "this regular expression could not be compiled: (assets"
+-- Left "This regular expression is malformed, please correct it:\n(assets"
 -- >>> _rsQuery $ set querystring ["assets"] defreportspec
 -- Acct (RegexpCI "assets")
 -- >>> _rsQuery $ set querystring ["(assets"] defreportspec
--- *** Exception: Updating ReportSpec failed: try using overEither instead of over or setEither instead of set
+-- *** Exception: Error: Updating ReportSpec failed: try using overEither instead of over or setEither instead of set
 -- >>> _rsQuery $ set period (MonthPeriod 2021 08) defreportspec
 -- Date DateSpan 2021-08
 class HasReportOptsNoUpdate a => HasReportOpts a where
@@ -809,6 +818,8 @@ class HasReportOptsNoUpdate a => HasReportOpts a where
     reportOpts = reportOptsNoUpdate
     {-# INLINE reportOpts #-}
 
+    -- XXX these names are a bit clashy
+
     period :: ReportableLens' a Period
     period = reportOpts.periodNoUpdate
     {-# INLINE period #-}
