@@ -1,19 +1,19 @@
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE TypeFamilies #-}
+{-# LANGUAGE TypeOperators #-}
 
 module Hledger.Utils.Parse (
   SimpleStringParser,
   SimpleTextParser,
   TextParser,
 
+  -- * SourcePos
   SourcePos(..),
   mkPos,
   unPos,
   initialPos,
-
-  -- * SourcePos
-  showSourcePosPair,
-  showSourcePos,
+  sourcePosPretty,
+  sourcePosPairPretty,
 
   choice',
   choiceInState,
@@ -38,41 +38,71 @@ module Hledger.Utils.Parse (
   skipNonNewlineSpaces1,
   skipNonNewlineSpaces',
 
+  -- ** Trace the state of hledger parsers
+  traceOrLogParse,
+  dbgparse,
+
   -- * re-exports
-  CustomErr
+  HledgerParseErrors,
+  HledgerParseErrorData,
+  customErrorBundlePretty,
 )
 where
 
+import Control.Monad (when)
+import qualified Data.Text as T
+import Text.Megaparsec
+import Text.Printf
 import Control.Monad.State.Strict (StateT, evalStateT)
 import Data.Char
 import Data.Functor (void)
 import Data.Functor.Identity (Identity(..))
 import Data.List
 import Data.Text (Text)
-import Text.Megaparsec
 import Text.Megaparsec.Char
 import Text.Megaparsec.Custom
-import Text.Printf
+import Hledger.Utils.Debug (debugLevel, traceOrLog)
 
 -- | A parser of string to some type.
-type SimpleStringParser a = Parsec CustomErr String a
+type SimpleStringParser a = Parsec HledgerParseErrorData String a
 
 -- | A parser of strict text to some type.
-type SimpleTextParser = Parsec CustomErr Text  -- XXX an "a" argument breaks the CsvRulesParser declaration somehow
+type SimpleTextParser = Parsec HledgerParseErrorData Text  -- XXX an "a" argument breaks the CsvRulesParser declaration somehow
 
 -- | A parser of text that runs in some monad.
-type TextParser m a = ParsecT CustomErr Text m a
+type TextParser m a = ParsecT HledgerParseErrorData Text m a
 
--- | Render source position in human-readable form.
-showSourcePos :: SourcePos -> String
-showSourcePos (SourcePos fp l c) =
-    show fp ++ " (line " ++ show (unPos l) ++ ", column " ++ show (unPos c) ++ ")"
+-- | Trace to stderr or log to debug log the provided label (if non-null)
+-- and current parser state (position and next input).
+-- See also: Hledger.Utils.Debug, megaparsec's dbg.
+-- Uses unsafePerformIO.
+traceOrLogParse :: String -> TextParser m ()
+traceOrLogParse msg = do
+  pos <- getSourcePos
+  next <- (T.take peeklength) `fmap` getInput
+  let (l,c) = (sourceLine pos, sourceColumn pos)
+      s  = printf "at line %2d col %2d: %s" (unPos l) (unPos c) (show next) :: String
+      s' = printf ("%-"++show (peeklength+30)++"s") s ++ " " ++ msg
+  traceOrLog s' $ return ()
+  where
+    peeklength = 30
 
--- | Render a pair of source position in human-readable form.
-showSourcePosPair :: (SourcePos, SourcePos) -> String
-showSourcePosPair (SourcePos fp l1 _, SourcePos _ l2 c2) =
-    show fp ++ " (lines " ++ show (unPos l1) ++ "-" ++ show l2' ++ ")"
-  where l2' = if unPos c2 == 1 then unPos l2 - 1 else unPos l2  -- might be at end of file withat last new-line
+-- class (Stream s, MonadPlus m) => MonadParsec e s m 
+-- dbgparse :: (MonadPlus m, MonadParsec e String m) => Int -> String -> m ()
+
+-- | Trace to stderr or log to debug log the provided label (if non-null)
+-- and current parser state (position and next input),
+-- if the global debug level is at or above the specified level.
+-- Uses unsafePerformIO.
+dbgparse :: Int -> String -> TextParser m ()
+dbgparse level msg = when (level <= debugLevel) $ traceOrLogParse msg
+
+-- | Render a pair of source positions in human-readable form, only displaying the range of lines.
+sourcePosPairPretty :: (SourcePos, SourcePos) -> String
+sourcePosPairPretty (SourcePos fp l1 _, SourcePos _ l2 c2) =
+    fp ++ ":" ++ show (unPos l1) ++ "-" ++ show l2'
+  where
+    l2' = if unPos c2 == 1 then unPos l2 - 1 else unPos l2  -- might be at end of file with a final new line
 
 -- | Backtracking choice, use this when alternatives share a prefix.
 -- Consumes no input if all choices fail.
@@ -81,7 +111,7 @@ choice' = choice . map try
 
 -- | Backtracking choice, use this when alternatives share a prefix.
 -- Consumes no input if all choices fail.
-choiceInState :: [StateT s (ParsecT CustomErr Text m) a] -> StateT s (ParsecT CustomErr Text m) a
+choiceInState :: [StateT s (ParsecT HledgerParseErrorData Text m) a] -> StateT s (ParsecT HledgerParseErrorData Text m) a
 choiceInState = choice . map try
 
 surroundedBy :: Applicative m => m openclose -> m a -> m a
@@ -92,7 +122,7 @@ parsewith p = runParser p ""
 
 -- | Run a text parser in the identity monad. See also: parseWithState.
 runTextParser, rtp
-  :: TextParser Identity a -> Text -> Either (ParseErrorBundle Text CustomErr) a
+  :: TextParser Identity a -> Text -> Either HledgerParseErrors a
 runTextParser = parsewith
 rtp = runTextParser
 
@@ -105,9 +135,9 @@ parsewithString p = runParser p ""
 parseWithState
   :: Monad m
   => st
-  -> StateT st (ParsecT CustomErr Text m) a
+  -> StateT st (ParsecT HledgerParseErrorData Text m) a
   -> Text
-  -> m (Either (ParseErrorBundle Text CustomErr) a)
+  -> m (Either HledgerParseErrors a)
 parseWithState ctx p = runParserT (evalStateT p ctx) ""
 
 parseWithState'
@@ -144,7 +174,7 @@ nonspace = satisfy (not . isSpace)
 isNonNewlineSpace :: Char -> Bool
 isNonNewlineSpace c = not (isNewline c) && isSpace c
 
-spacenonewline :: (Stream s, Char ~ Token s) => ParsecT CustomErr s m Char
+spacenonewline :: (Stream s, Char ~ Token s) => ParsecT HledgerParseErrorData s m Char
 spacenonewline = satisfy isNonNewlineSpace
 {-# INLINABLE spacenonewline #-}
 
@@ -152,20 +182,19 @@ restofline :: TextParser m String
 restofline = anySingle `manyTill` eolof
 
 -- Skip many non-newline spaces.
-skipNonNewlineSpaces :: (Stream s, Token s ~ Char) => ParsecT CustomErr s m ()
-skipNonNewlineSpaces = () <$ takeWhileP Nothing isNonNewlineSpace
+skipNonNewlineSpaces :: (Stream s, Token s ~ Char) => ParsecT HledgerParseErrorData s m ()
+skipNonNewlineSpaces = void $ takeWhileP Nothing isNonNewlineSpace
 {-# INLINABLE skipNonNewlineSpaces #-}
 
 -- Skip many non-newline spaces, failing if there are none.
-skipNonNewlineSpaces1 :: (Stream s, Token s ~ Char) => ParsecT CustomErr s m ()
-skipNonNewlineSpaces1 = () <$ takeWhile1P Nothing isNonNewlineSpace
+skipNonNewlineSpaces1 :: (Stream s, Token s ~ Char) => ParsecT HledgerParseErrorData s m ()
+skipNonNewlineSpaces1 = void $ takeWhile1P Nothing isNonNewlineSpace
 {-# INLINABLE skipNonNewlineSpaces1 #-}
 
 -- Skip many non-newline spaces, returning True if any have been skipped.
-skipNonNewlineSpaces' :: (Stream s, Token s ~ Char) => ParsecT CustomErr s m Bool
+skipNonNewlineSpaces' :: (Stream s, Token s ~ Char) => ParsecT HledgerParseErrorData s m Bool
 skipNonNewlineSpaces' = True <$ skipNonNewlineSpaces1 <|> pure False
 {-# INLINABLE skipNonNewlineSpaces' #-}
 
-
 eolof :: TextParser m ()
 eolof = void newline <|> eof
