@@ -10,6 +10,7 @@ module Hledger.Utils.String (
  -- quoting
  quoteIfNeeded,
  singleQuoteIfNeeded,
+ quoteForCommandLine,
  -- quotechars,
  -- whitespacechars,
  words',
@@ -19,6 +20,9 @@ module Hledger.Utils.String (
  strip,
  lstrip,
  rstrip,
+ strip1Char,
+ stripBy,
+ strip1By,
  chomp,
  chomp1,
  singleline,
@@ -34,7 +38,7 @@ module Hledger.Utils.String (
 
 
 import Data.Char (isSpace, toLower, toUpper)
-import Data.List (intercalate)
+import Data.List (intercalate, dropWhileEnd)
 import qualified Data.Text as T
 import Text.Megaparsec ((<|>), between, many, noneOf, sepBy)
 import Text.Megaparsec.Char (char)
@@ -68,6 +72,25 @@ lstrip = dropWhile isSpace
 rstrip :: String -> String
 rstrip = reverse . lstrip . reverse
 
+-- | Strip the given starting and ending character
+-- from the start and end of a string if both are present.
+strip1Char :: Char -> Char -> String -> String
+strip1Char b e s = case s of
+  (c:cs) | c==b, not $ null cs, last cs==e -> init cs
+  _ -> s
+
+-- | Strip a run of zero or more characters matching the predicate
+-- from the start and end of a string.
+stripBy :: (Char -> Bool) -> String -> String
+stripBy f = dropWhileEnd f . dropWhile f
+
+-- | Strip a single balanced enclosing pair of a character matching the predicate
+-- from the start and end of a string.
+strip1By :: (Char -> Bool) -> String -> String
+strip1By f s = case s of
+  (c:cs) | f c, not $ null cs, last cs==c -> init cs
+  _ -> s
+
 -- | Remove all trailing newlines/carriage returns.
 chomp :: String -> String
 chomp = reverse . dropWhile (`elem` "\r\n") . reverse
@@ -118,15 +141,44 @@ quoteIfNeeded s | any (`elem` s) (quotechars++whitespacechars++redirectchars) =
     escapeQuotes (c:cs)   x = showChar c        $ escapeQuotes cs x
 
 -- | Single-quote this string if it contains whitespace or double-quotes.
--- No good for strings containing single quotes.
+-- Does not work for strings containing single quotes.
 singleQuoteIfNeeded :: String -> String
-singleQuoteIfNeeded s | any (`elem` s) (quotechars++whitespacechars) = "'"++s++"'"
+singleQuoteIfNeeded s | any (`elem` s) (quotechars++whitespacechars) = singleQuote s
                       | otherwise = s
 
-quotechars, whitespacechars, redirectchars :: [Char]
+-- | Prepend and append single quotes to a string.
+singleQuote :: String -> String
+singleQuote s = "'"++s++"'"
+
+-- | Try to single- and backslash-quote a string as needed to make it usable
+-- as an argument on a (sh/bash) shell command line. At least, well enough 
+-- to handle common currency symbols, like $. Probably broken in many ways.
+--
+-- >>> quoteForCommandLine "a"
+-- "a"
+-- >>> quoteForCommandLine "\""
+-- "'\"'"
+-- >>> quoteForCommandLine "$"
+-- "'\\$'"
+--
+quoteForCommandLine :: String -> String
+quoteForCommandLine s
+  | any (`elem` s) (quotechars++whitespacechars++shellchars) = singleQuote $ quoteShellChars s
+  | otherwise = s
+
+-- | Try to backslash-quote common shell-significant characters in this string.
+-- Doesn't handle single quotes, & probably others.
+quoteShellChars :: String -> String
+quoteShellChars = concatMap escapeShellChar
+  where
+    escapeShellChar c | c `elem` shellchars = ['\\',c]
+    escapeShellChar c = [c]
+
+quotechars, whitespacechars, redirectchars, shellchars :: [Char]
 quotechars      = "'\""
 whitespacechars = " \t\n\r"
 redirectchars   = "<>"
+shellchars      = "<>(){}[]$7?#!~`"
 
 -- | Quote-aware version of words - don't split on spaces which are inside quotes.
 -- NB correctly handles "a'b" but not "''a''". Can raise an error if parsing fails.
